#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import numpy as np
from abc import ABC, abstractmethod

from ads.common import logger, utils
from ads.explanations.base_explainer import GlobalExplainer
from ads.explanations.mlx_interface import check_tabular_or_text
from ads.explanations.mlx_interface import init_lime_explainer
from ads.explanations.mlx_interface import init_permutation_importance_explainer
from ads.explanations.mlx_interface import (
    init_partial_dependence_explainer,
    init_ale_explainer,
)
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common.decorator.deprecate import deprecated


class MLXGlobalExplainer(GlobalExplainer):
    """
    Global Explainer class.

    Generates global explanations to help understand the general model
    behavior. Supported explanations:

        - (Tabular) Feature Permutation Importance.
        - (Tabular) Partial Dependence Plots (PDP) & Individual Conditional
          Expectation (ICE).
        - (Text) Aggregate local explanations (global explanation approximation
          constructed from multiple local explanations).

    Supports:

        - Binary classification.
        - Multi-class classification.
        - Regression.

    """

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self):
        super(GlobalExplainer, self).__init__()
        self.explainer = None
        self.selected_features = None
        self.pdp_explainer = None
        self.ale_explainer = None

    def compute_feature_importance(
        self,
        n_iter=20,
        sampling=None,
        balance=False,
        scoring_metric=None,
        selected_features=None,
    ):
        """
        Generates a global explanation to help understand the general behavior
        of the model. This explainer identifies which features are most important
        to the model.

        If the dataset is tabular, computes a global feature permutation importance
        explanation. If the dataset is text, approximates a global explanation by
        generating and aggregating multiple local explanations.

        Parameters
        ----------
        n_iter : int, optional
            Number of iterations of the permutation importance algorithm to
            perform. Increasing this value increases the quality/stability of
            the explanation, but increases the explanation time. Default value is 20.
        sampling : dict, optional
            If not `None`, the dataset is clustered or sampled according to the
            provided technique. `sampling` is a dictionary containing the technique
            to use and the corresponding parameters. Format is described below:

            - `technique`: Either `cluster` or `random`.
            - If `cluster`, also requires:

                - `eps`: Maximum distance between two samples to be considered
                  in the same cluster.
                - `min_samples`: Minimum number of samples to include in each
                  cluster.

            - If `random`, also requires:

                - `n_samples`: Number of samples to return.

            By default None. Note that text datasets are always sampled. If not provided
            with a sampling option, defaults to 40 random samples.
        balance : bool, optional
            If True, the dataset will be balanced via sampling. If 'sampling' is not
            set, the sampling technique defaults to 'random'.
        scoring_metric : string, optional
            If specified, propegates a string indicating the supported scoring metric.
            The scoring metrics available out of the box are the ones made available
            by ScyPy. Supported Metrics:

            - Multi-class Classification

                `f1_weighted`, `f1_micro`, `f1_macro`, `recall_weighted`, `recall_micro`, `recall_macro`,
                `accuracy`, `balanced_accuracy`, `roc_auc`, `precision_weighted`, `precision_macro`,
                `precision_micro`

            - Binary Classification

                Same as multi-class classification

            - Regression

                `r2`, `neg_mean_squared_error`, `neg_root_mean_squared_error`, `neg_mean_absolute_error`,
                `neg_median_absolute_error`, `neg_mean_absolute_percentage_error`,
                `neg_symmetric_mean_absolute_percentage_error`
        selected_features: list[str], list[int], optional
            List of the selected features. It can be any subset of
            the original features that are in the dataset provided to the model.
            Default value is None.

        Returns
        -------
        :class:FeatureImportance
            `FeaturePermutationImportance` explanation object.

        """
        self.selected_features = selected_features
        self.configure_feature_importance(selected_features=self.selected_features)
        if self.explainer.config.type == "text":
            labels = list(range(len(self.class_names)))
            # The requirement to downsample the text datasets should be fixed at somepoint
            if sampling is None:
                sampling = {"technique": "random", "n_samples": 40}
            explanation = self.explainer.explain_aggregate_local(
                self.X_test, sampling=sampling, labels=labels
            )
        else:
            if self.mode_ == "regression":
                allowed_metrics = [
                    "r2",
                    "neg_mean_squared_error",
                    "neg_root_mean_squared_error",
                    "neg_mean_absolute_error",
                    "neg_median_absolute_error",
                    "neg_mean_absolute_percentage_error",
                    "neg_symmetric_mean_absolute_percentage_error",
                ]
            elif self.mode_ == "classification" and len(self.class_names) == 2:
                # Binary classification
                allowed_metrics = [
                    "f1_weighted",
                    "f1_micro",
                    "f1_macro",
                    "recall_weighted",
                    "recall_micro",
                    "recall_macro",
                    "accuracy",
                    "balanced_accuracy",
                    "roc_auc",
                    "precision_weighted",
                    "precision_macro",
                    "precision_micro",
                ]
            else:
                # Multiclass classification
                allowed_metrics = [
                    "f1_weighted",
                    "f1_micro",
                    "f1_macro",
                    "recall_weighted",
                    "recall_micro",
                    "recall_macro",
                    "accuracy",
                    "balanced_accuracy",
                    "roc_auc",
                    "precision_weighted",
                    "precision_macro",
                    "precision_micro",
                ]
            if scoring_metric not in allowed_metrics and scoring_metric is not None:
                raise Exception(
                    "Scoring Metric not supported for this type of problem: {}, for problem type {}, the availble supported metrics are {}".format(
                        scoring_metric, self.mode_, allowed_metrics
                    )
                )
            if balance and sampling is None:
                sampling = {"technique": "random"}
            try:
                explanation = self.explainer.compute(
                    self.X_test,
                    self.y_test,
                    n_iter=n_iter,
                    sampling=sampling,
                    balance=balance,
                    scoring_metric=scoring_metric,
                )
            except IndexError as e:
                if selected_features is not None:
                    raise IndexError(
                        f"Unable to calculate permutation importance due to: {e}. "
                        f"selected_features must be a list of features within the bounds of the existing features "
                        f"(that were provided to model). Provided selected_features: {selected_features}."
                    )
            except Exception as e:
                logger.error(
                    f"Unable to calculate permutation importance scores due to: {e}."
                )
                raise e
        return FeatureImportance(explanation, self.class_names, self.explainer.config)

    def compute_partial_dependence(
        self, features, partial_range=(0.00, 1.0), num_samples=30, sampling=None
    ):
        """
        Generates a global partial dependence plot (PDP) and individual conditional
        expectation (ICE) plots to help understand the relationship between feature
        values and the model target.

        Only supported for tabular datasets.

        Parameters
        ----------

        features : list of int, list of str
            List of feature names or feature indices to explain.
        partial_range : tuple, optional
            2-tuple with the minimum and maximum percentile values to consider for the PDP from
            the feature's train distribution. Must be between 0.0 and 1.0.
            Defaults to `partial_range = (0.05, 0.95)`.
        num_samples : int, optional
            Maximum number of samples to generate for each feature within the
            `partial_range` of its value distribution. Increasing this value
            generates more points to evaluate, but increases the explanation
            time. If there are fewer unique values for a feature within the
            `partial_range`, the number of unique values is selected. For two-feature
            PDP, the total number of evaluated samples is the multiplication
            of `num_samples`. Default value is 30.
        sampling : dict, optional
            If not None, the dataset will be clustered or sampled according to the
            provided technique. 'sampling' is a dictionary containing the technique
            to use and the corresponding parameters. Format is described below:

            - `technique`: Either "cluster" or "random".
            - `cluster` also requires:

                - `eps`: Maximum distance between two samples to be considered
                  in the same cluster.
                - `min_samples`: Minimum number of samples to include in each
                  cluster.

            - `random` also requires:

                - 'n_samples': Number of samples to return.

            Default value is `None` (no sampling).

        Returns
        -------
        :class:MLXPartialDependencies
            `MLXPartialDependencies` object.

        """
        if self.pdp_explainer is None:
            self._init_partial_dependence()

        # Wrap in a list if a list is not provided
        if not isinstance(features, list):
            features = [features]

        # Convert to uppercase to be case-insensitive
        features = [str(f).upper() for f in features]
        feature_names = np.char.upper(self.X_train.columns.tolist())

        # Fail if we were not provided valid feature names
        if not all(np.isin(features, feature_names)):
            print("One or more features (%s) does not exist in data." % str(features))
            print("Existing features: %s" % str(feature_names))
            return

        # Extract the feature ids
        feature_ids = np.where(np.isin(feature_names, features))[0].tolist()

        if check_tabular_or_text(self.est, self.X_train) == "tabular":
            if len(feature_ids) > 2:
                raise ValueError("Maximum number of partial dependency features is 2.")

            return MLXPartialDependencies(
                pdp=self.pdp_explainer.compute(
                    data=self.X_train,
                    partial_ids=feature_ids,
                    partial_range=partial_range,
                    num_samples=num_samples,
                    sampling=sampling,
                ),
                pdp_exp=self.pdp_explainer,
            )
        else:
            raise ValueError(
                "Partial Dependence Plot is not supported for text classification dataset."
            )

    def compute_accumulated_local_effects(
        self,
        feature,
        partial_range=(0.00, 1.0),
        num_samples=30,
        sampling=None,
        corr_threshold=0.7,
    ):
        """
        Generates the accumulated local effects plots to help understand the relationship between feature
        values and the model target.

        Only supported for tabular datasets.

        Parameters
        ----------
        feature : str
            Feature name to explain.
        partial_range : tuple, optional
            Min/max percentile values to consider for the ALE from the
            feature's train distribution. Must be between 0.0 and 1.0.
            By default `partial = (0.05, 0.95)`.
        num_samples : int, optional
            Maximum number of samples to generate for each feature within the
            `partial_range` of its value distribution. Increasing this value
            generates more points to evaluate, but increases the explanation
            time. If there are fewer unique values for a feature within the
            `partial_range`, the number of unique values is selected.
        sampling : dict, optional
            If not `None`, the dataset is clustered or sampled according to the
            provided technique. `sampling` is a dictionary containing the technique
            to use and the corresponding parameters. The format is:

            - `technique`: Can be either "cluster" or "random".
            - `cluster` also requires:

                - `eps`: Maximum distance between two samples to be considered
                  in the same cluster.
                - `min_samples`: Minimum number of samples to include in each
                  cluster.

            - `random` also requires:

                - `n_samples`: Number of samples to return.

            Defaults to `None` (no sampling).

        corr_threshold : float, optional
             Value between 0.0 and 1.0 for which a feature is considered highly correlated with
             another feature (Default = 0.7).

        Returns
        -------
        :class:MLXAccumulatedLocalEffects
            `AccumulatedLocalEffects` explanation object.

        """
        if self.ale_explainer is None:
            self._init_accumulated_local_effects()

        # Wrap in a list if a list is not provide, to be able support list of two features in the near future.
        if not isinstance(feature, list):
            feature = [feature]

        # Convert to uppercase to be case-insensitive
        feature = [str(f).upper() for f in feature]
        feature_names = np.char.upper(self.X_train.columns.tolist())

        # Fail if we were not provided valid feature names
        if not all(np.isin(feature, feature_names)):
            print("One or more features (%s) does not exist in data." % str(feature))
            print("Existing features: %s" % str(feature_names))
            return

        # Extract the feature ids
        feature_ids = np.where(np.isin(feature_names, feature))[0].tolist()

        if check_tabular_or_text(self.est, self.X_train) == "tabular":
            if len(feature_ids) > 1:
                raise ValueError(
                    "Maximum number of Accumulated Local Effects features is 1."
                )

            return MLXAccumulatedLocalEffects(
                ale=self.ale_explainer.compute(
                    data=self.X_train,
                    partial_ids=feature_ids,
                    partial_range=partial_range,
                    num_samples=num_samples,
                    sampling=sampling,
                    corr_threshold=corr_threshold,
                ),
                ale_exp=self.ale_explainer,
            )
        else:
            raise ValueError(
                "Accumulated Local Effects Plot is not supported for text classification dataset."
            )

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def show_in_notebook(self):  # pragma: no cover
        """
        Generates and visualizes the global feature importance explanation.
        """
        with utils.get_progress_bar(3, description="Model Explanation") as bar:
            bar.update("begin computing")
            bar.update("calculating feature importance")
            explainer_holder = self.compute_feature_importance(
                selected_features=self.selected_features
            )
            plot1 = explainer_holder.show_in_notebook()
            bar.update("calculating partial dependence plot")
            pdp_plot_feature_name = explainer_holder.explanation
            # pdp_plot_feature_name = explainer_holder.explanation.get_global_explanation().index[0]
            pdp_plot = self.compute_partial_dependence([pdp_plot_feature_name])
            # plot2 = pdp_plot.show_in_notebook()

        from IPython.core.display import display, HTML

        display(HTML(plot1.data))
        # display(HTML(plot1.data + plot2.data))

    def configure_feature_importance(self, **kwargs):
        """
        Validates and initializes the feature importance explainer based on the provided
        configuration parameters in kwargs. Tabular datasets use the feature permutation
        importance explainer, text datasets use the aggregate local explainer.

        Supported configuration options:

        - For tabular datasets:

            - `client`: Currently only allowed to be None to disable parallelization.
            - `random_state`: None, int, or instance of Randomstate.
            - `selected_features`: None, or list of the selected features.

        - For text datasets:

            - `surrogate_model`: Surrogate model to use. Can be 'linear' or 'decision_tree'.
            - `num_samples`: Number of generated samples to fit the surrogate model. Int.
            - `exp_sorting`: Feature importance sorting. Can be 'absolute' or 'ordered'.
            - `scale_weight`: Normalizes the feature importance coefficients from LIME to sum to one.
            - `client`: Currently only allowed to be None to disable parallelization.
            - `batch_size`: Number of local explanations per Dask worker.
            - `random_state`: None, int, or instance of Randomstate.
            - `selected_features`: None, or list of the selected features.

        Parameters
        ----------
        kwargs : dict
            Keyword parameter dictionary.

        Returns
        -------
        MLXGlobalExplainer
            the modified instance (self)

        """

        if check_tabular_or_text(self.est, self.X_train) == "tabular":
            avail_args = ["client", "random_state", "selected_features"]
        else:
            avail_args = [
                "client",
                "random_state",
                "surrogate_model",
                "num_samples",
                "exp_sorting",
                "scale_weight",
                "batch_size",
                "selected_features",
            ]

        for k, _ in kwargs.items():
            if k not in avail_args:
                raise ValueError(
                    "Unexpected argument for the feature importance explainer: {}".format(
                        k
                    )
                )

        if kwargs.get("client", None) is not None:
            raise ValueError(
                "Invalid client provided. Currently only supports disabling parallelization "
                "by setting client=None"
            )
        if kwargs.get("surrogate_model", None) not in ["linear", "decision_tree", None]:
            raise ValueError(
                "Invalid surrogate_model provided. Currently only supports linear or decision_tree"
            )
        selected_features = kwargs.get("selected_features")
        if selected_features is not None and not isinstance(selected_features, list):
            raise ValueError(
                f"selected_features ({selected_features}) value must be a list of features, "
                f"but it is of type: {type(selected_features)}."
            )

        self._init_feature_importance(**kwargs)
        return self

    def configure_partial_dependence(self, **kwargs):
        """
        Validates and initializes the partial dependence explainer based on the provided
        configuration parameters in kwargs. Only supports tabular datasets.

        Supported configuration options:
            client: Currently only supports 'None' to disable parallelization.

        Parameters
        ----------
        kwargs : dict
            Keyword parameter dictionary.

        Returns
        -------
        MLXGlobalExplainer
            the modified instance (self)
        """

        for k, _ in kwargs.items():
            if k not in ["client"]:
                raise ValueError(
                    "Unexpected argument for the partial dependence explainer: {}".format(
                        k
                    )
                )
        if kwargs.get("client", None) is not None:
            raise ValueError(
                "Invalid client provided. Currently only supports disabling parallelization "
                "by setting client=None"
            )
        self._init_partial_dependence(**kwargs)
        return self

    def configure_accumulated_local_effects(self, **kwargs):
        """
        Validates and initializes the accumulated local effects explainer based on the provided
        configuration parameters in kwargs. Only supports tabular datasets.

        Supported configuration options:

        - client: Currently only supports 'None' to disable parallelization.

        Parameters
        ----------
        kwargs : dict
            Keyword parameter dictionary.

        Returns
        -------
        MLXGlobalExplainer
            the modified instance (self)
        """

        for k, _ in kwargs.items():
            if k not in ["client"]:
                raise ValueError(
                    "Unexpected argument for the accumulated local effects explainer: {}".format(
                        k
                    )
                )
        if kwargs.get("client", None) is not None:
            raise ValueError(
                "Invalid client provided. Currently only supports disabling parallelization "
                "by setting client=None"
            )
        self._init_accumulated_local_effects(**kwargs)
        return self

    def feature_importance_summary(self):
        """
        Displays detailed information about the feature importance explainer.

        Returns
        -------
        str
            HTML object representing the explainer summary.

        """

        if self.explainer is None:
            self.compute_feature_importance(selected_features=self.selected_features)
        return self.explainer.show_in_notebook()

    def partial_dependence_summary(self):
        """
        Displays detailed information about the partial dependence explainer.

        Returns
        -------
        str
            HTML object representing the explainer summary.
        """

        if self.pdp_explainer is None:
            self._init_partial_dependence()
        return self.pdp_explainer.show_in_notebook()

    def accumulated_local_effects_summary(self):
        """
        Displays detailed information about the accumulated local effects explainer.

        Returns
        -------
        str
            HTML object representing the explainer summary.
        """

        if self.ale_explainer is None:
            self._init_accumulated_local_effects()
        return self.ale_explainer.show_in_notebook()

    def _init_feature_importance(self, **kwargs):
        """
        Internal function to initialize the feature importance explainer. Tabular datasets
        use the feature permutation importance explainer, text datasets use the aggregate local
        explainer.

        Parameters
        ----------
        kwargs : dict
            Keyword parameter dictionary.
        """
        if self.mode == "regression":
            self.class_names_ = ["Target"]
        if check_tabular_or_text(self.est, self.X_train) == "tabular":
            self.explainer = init_permutation_importance_explainer(
                self.explainer,
                self.est,
                self.X_train,
                self.y_train,
                self.mode,
                class_names=self.class_names,
                **kwargs,
            )
        else:
            self.explainer = init_lime_explainer(
                self.explainer,
                self.est,
                self.X_train,
                self.y_train,
                self.mode,
                class_names=self.class_names,
                **kwargs,
            )

    def _init_partial_dependence(self, **kwargs):
        """
        Internal function to initialize the partial dependence explainer.

        Parameters
        ----------
        kwargs : dict
            Keyword parameter dictionary.
        """
        if self.mode == "regression":
            self.class_names_ = ["Target"]
        self.pdp_explainer = init_partial_dependence_explainer(
            self.pdp_explainer,
            self.est,
            self.X_train,
            self.y_train,
            self.mode,
            class_names=self.class_names,
            **kwargs,
        )

    def _init_accumulated_local_effects(self, **kwargs):
        """
        Internal function to initialize the accumulated local effects explainer.

        Parameters
        ----------
        kwargs : dict
            Keyword parameter dictionary.
        """
        if self.mode == "regression":
            self.class_names_ = ["Target"]
        self.ale_explainer = init_ale_explainer(
            self.ale_explainer,
            self.est,
            self.X_train,
            self.y_train,
            self.mode,
            class_names=self.class_names,
            **kwargs,
        )


class MLXFeatureDependenceExplanation(ABC):

    __name__ = "MLXFeatureDependenceExplanation"

    def __init__(self, fd, fd_exp):
        self.fd = fd
        self.fd_exp = fd_exp

    @abstractmethod
    def show_in_notebook(
        self,
        labels=None,
        cscale="YIGnBu",
        show_distribution=True,
        discrete_threshold=0.15,
        # line_gap=0,  # will add it back after ALE starts handling two features, remember to add the doc string too
        show_correlation_warning=True,
        centered=False,
        show_median=True,
    ):  # pragma: no cover
        """
        Visualize PDP/ICE plots in the Notebook.

        Parameters
        ----------
        labels : tuple, list, int, bool, str, optional
            labels to visualize.
        cscale : str, optional
            Plotly color scale to use for the heatmap. See the standard Plotly color scales for available options
            Default value is "YIGnBu".
        show_distribution : bool, optional
            If `True`, the feature’s value distribution (from the train set) will be shown along the
            corresponding axis in the 1-feature or 2-feature plot. Default is `True`.
        discrete_threshold : float, optional
            Value between 0.0 and 1.0 indicating the fraction of unique values required for a numerical feature
            to be considered discrete or continuous. Default is 0.15.
        show_correlation_warning : bool, optional
            If `True`, the correlated feature warning is shown. Default is `True`.
        centered : bool, optional
            If `True`, ICE plots is centered based on the first value of each sample (i.e., all values are
            subtracted from the first value). Default is False.
        show_median : bool, optional
            If True, a median line is included in the ICE explanation plot. Default is True.

        Returns
        -------
        str
            Plotly HTML object containing a line chart, heat map, or violin plot for this feature dependence explanation
        """
        pass

    def as_dataframe(self):
        """
        Returns the raw explanation data as a pandas.DataFrame.

        Returns
        -------
        pandas.DataFrame
            DataFrame containing the raw PDP explanation data.
        """
        return self.fd.as_dataframe()

    def get_diagnostics(self):
        """
        Extracts the raw explanation and evaluation data from the explanation object
        (Used to generate the visualizations).

        Returns
        -------
        dict
            Dictionary containing the raw explanation/evaluation data.
        """
        return self.fd.get_diagnostic()


class MLXPartialDependencies(MLXFeatureDependenceExplanation):
    """
    Represents the object constructed by the :class:`MLXGlobalExplainer`.

    Contains functions to visualize the explanation and extract raw explanation data.
    """

    __name__ = "MLXPartialDependencies"

    def __init__(self, pdp, pdp_exp):
        super(MLXPartialDependencies, self).__init__(pdp, pdp_exp)

    def show_in_notebook(
        self,
        mode="pdp",
        labels=None,
        cscale="YIGnBu",
        show_distribution=True,
        discrete_threshold=0.15,
        line_gap=0,
        show_correlation_warning=True,
        centered=False,
        show_median=True,
    ):
        """
        Visualize PDP/ICE plots in the Notebook.

        Parameters
        ----------
        mode : str, optional
            Type to visualize. Either "pdp" or "ice". Default is "pdp".
        labels : tuple, list, int, bool, str, optional
            labels to visualize.
        cscale : str, optional
            Plotly color scale to use for the heatmap. See the standard Plotly color scales for available options
            Default value is "YIGnBu".
        show_distribution : bool, optional
            If `True`, the feature’s value distribution (from the train set) will be shown along the
            corresponding axis in the 1-feature or 2-feature plot. Default is `True`.
        discrete_threshold : float, optional
            Value between 0.0 and 1.0 indicating the fraction of unique values required for a numerical feature
            to be considered discrete or continuous. Default is 0.15.
        line_gap : int, optional
            Width of the gap between values in the two-feature PDP heat map. Default is 0.
        show_correlation_warning : bool, optional
            If `True`, the correlated feature warning are shown. Default is `True`.
        centered : bool, optional
            If `True`, ICE plots are centered based on the first value of each sample (i.e., all values are
            subtracted from the first value). Default is `False`.
        show_median : bool, optional
            If `True`, a median line is included in the ICE explanation plot. Default is `True`.

        Returns
        -------
        str
            Plotly HTML object containing a line chart, heat map, or violin plot for this feature dependence explanation
        """
        return self.fd.show_in_notebook(
            mode=mode,
            labels=labels,
            cscale=cscale,
            show_distribution=show_distribution,
            discrete_threshold=discrete_threshold,
            line_gap=line_gap,
            show_correlation_warning=show_correlation_warning,
            centered=centered,
            show_median=show_median,
        )


class MLXAccumulatedLocalEffects(MLXFeatureDependenceExplanation):
    """
    Accumulated Local Effects explanation object constructed by the `:class:MLXGlobalExplainer`.

    Contains functions to visualize the explanation in a Notebook and extract the
    raw explanation data.
    """

    __name__ = "MLXAccumulatedLocalEffects"

    def __init__(self, ale, ale_exp):
        super(MLXAccumulatedLocalEffects, self).__init__(ale, ale_exp)

    def show_in_notebook(
        self,
        labels=None,
        cscale="YIGnBu",
        show_distribution=True,
        discrete_threshold=0.15,
        show_correlation_warning=True,
        centered=False,
        show_median=True,
    ):
        """
        Visualize ALE plots in the Notebook.

        Parameters
        ----------
        labels : tuple, list, int, bool, str, optional
            labels to visualize.
        cscale : str, optional
            Plotly color scale to use for the heatmap. See the standard Plotly color scales for available options
            Default value is "YIGnBu".
        show_distribution : bool, optional
            If `True`, the feature’s value distribution (from the train set) will be shown along the
            corresponding axis in the 1-feature. Default is `True`.
        discrete_threshold : float, optional
            Value between 0.0 and 1.0 indicating the fraction of unique values required for a numerical feature
            to be considered discrete or continuous. Default is 0.15.
        show_correlation_warning : bool, optional
            If `True`, the correlated feature warning will be shown. Default is `True`.
        centered : bool, optional
            If `True`, ALE plots will be centered based on the first value of each sample (i.e., all values are
            subtracted from the first value). Default is `False`.
        show_median : bool, optional
            If `True`, a median line is included in the ALE explanation plot. Default is `True`.

        Returns
        -------
        str
            Plotly HTML object containing a line chart, heat map, or violin plot for this feature dependence explanation
        """
        return self.fd.show_in_notebook(
            mode="pdp",
            labels=labels,
            cscale=cscale,
            show_distribution=show_distribution,
            discrete_threshold=discrete_threshold,
            line_gap=0,
            show_correlation_warning=show_correlation_warning,
            centered=centered,
            show_median=show_median,
        )


class FeatureImportance:
    """
    Feature Permutation Importance explanation object constructed by the
    :class:`MLXGlobalExplainer` class.

    Contains functions to visualize the explanation in a Notebook and extract the
    raw explanation data.
    """

    def __init__(self, explanation, class_names, type):
        self.explanation = explanation
        self.class_names = class_names
        self.type = type
        if isinstance(self.class_names, np.ndarray):
            self.class_names = self.class_names.tolist()

    def show_in_notebook(
        self,
        mode=None,
        show=None,
        labels=None,
        cscale="YIGnBu",
        colormap=None,
        return_wordcloud=False,
        n_features=None,
        **kwargs,
    ):
        """
        Generates a visualization for the local explanation. Depending on the type of explanation, different
        visualizations are supported. See the "mode" and "show" parameters below.

        Parameters
        ----------
        mode : str
            Type of visualization to generate. Certain visualization modes are only supported for either text or
            tabular datasets. Supported options:

            - `bar`: Generates a horizontal bar chart for the most important features (text and tabular).
            - `stacked`: Generates a stacked horizontal bar chart for the most important features (text and tabular).
            - `box_plot`: Generates a box plot for the most important features, which provides more information
              about the explanation over the different iterations of the feature permutation importance
              algorithm (tabular only).
            - `detailed`: Generates a scatter plot for the most important features, providing even more
              information about the explanation over the different iterations of the Feature Permutation
              Importance algorithm (tabular only).
            - `heatmap`: Generates a heatmap representing the average feature/word importance over multiple
              local explanations (aggregates local explanations). Average feature importance is measured by
              the fraction of local explanations where a given feature was assigned a given importance
              (text only).
            - `wordcloud`: Generates a wordcloud from the average feature importance. Features/words with
              higher importance are larger than features/words with lower importance (text only).

            Default value is "bar" for tabular and "wordcloud" for text.
        show : str
            (text only) Secondary visualization mode for configuring the visualization.
            Can be one of:

            - `absolute`: The absolute value of feature importances are shown (i.e., a feature that is highly
              important towards or against the target label is considered important).
            - `posneg`: Shows both the positive and negative global feature attributions. For bar, the features
              are ordered based on their absolute feature importance (sum of pos/neg) and a dual bar chart shows
              the fraction of local explanations where the feature contributed both towards and against the
              corresponding label. For wordcloud, two wordclouds are generated for the positive and negative feature
              importances. Only valid for mode=bar and mode=wordcloud. `mode=heatmap` defaults to `show=absolute`.

        labels : tuple, list, int, bool, str
            (text only) Label indices to visualize. If `None`, all of the labels that the explanation was generated for
            will be visualized. By default None.
        cscale : str, optional
            Plotly color scale to use for the heatmap. See the standard Plotly color scales for available
            options. Default value is "YIGnBu".
        colormap : list of str, optional
            List of colormaps to use for the wordclouds. One per label. Defaults to `None`.
        return_wordcloud : bool, optional
            If `True`, the generated wordcloud objects are returned instead of visualized. Defaults
            to `False`.
        n_features : int, optional
            (tabular only). Allows the user to visualize a subset of the top-N most important features from the explainer.
            If `n_features` is `None` or greater than the total number of features, all features are shown. If
            `n_features` is not an `int` or <= 0, an exception is thrown.
        kwargs : dict
            Keyword arguments for configuring the wordclouds.

        Returns
        -------
        str, list of wordcloud
            HTML string for the visualization or list of generated wordcloud objects if `return_wordcloud=True`,
            two per label (+/-).
        """
        if self.type.type == "text":
            if labels:
                labels = [
                    self.class_names.index(label) if isinstance(label, str) else label
                    for label in labels
                ]
            else:
                labels = list(range(len(self.class_names)))
            if len(self.class_names) == 2:
                return self.explanation.show_in_notebook()
            else:
                return self.explanation.show_in_notebook(
                    mode=mode if mode else "wordcloud",
                    show=show,
                    labels=labels,
                    cscale=cscale,
                    colormap=colormap,
                    return_wordcloud=return_wordcloud,
                    **kwargs,
                )
        else:
            if labels:
                raise ValueError("label is supported only for text explanation.")
            return self.explanation.show_in_notebook(
                n_features=n_features, mode=mode if mode else "bar"
            )

    def get_global_explanation(self):
        """
        Returns the raw global explanation data only.

        Returns
        -------
        dict
            Dictionary containing raw explanation data.
        """
        return self.explanation.get_global_explanation()

    def get_diagnostics(self):
        """
        Extracts the raw explanation and evaluation data from the explanation object
        (Used to generate the visualizations).

        Returns
        -------
        dict
            Dictionary containing the raw explanation/evaluation data.
        """
        return self.explanation.get_diagnostic()


class GlobalExplanationsException(TypeError):
    def __init__(self, msg):
        super(GlobalExplanationsException, self).__init__(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
import logging
from typing import Dict, Union

import numpy as np
import pandas as pd

pd.options.mode.chained_assignment = None


# Note to developers: If you make any changes to this class, copy and paste those changes over to
# templates/score_onnx.jinja2 and templates/score_onnx_new.jinja2. We do not yet have an automatic way of doing this.
class ONNXTransformer(object):
    """
    This is a transformer to convert X [pandas.Dataframe, pd.Series] data into Onnx
    readable dtypes and formats. It is Serializable, so it can be reloaded at another time.


    Examples
    --------
    >>> from ads.model.transformer.onnx_transformer import ONNXTransformer
    >>> onnx_data_transformer = ONNXTransformer()
    >>> train_transformed = onnx_data_transformer.fit_transform(train.X, {"column_name1": "impute_value1", "column_name2": "impute_value2"}})
    >>> test_transformed = onnx_data_transformer.transform(test.X)
    """

    def __init__(self):
        self.impute_values = {}
        self.dtypes = None
        self._fitted = False

    @staticmethod
    def _handle_dtypes(X: Union[pd.DataFrame, pd.Series, np.ndarray, list]):
        """Handles the dtypes for pandas dataframe and pandas Series.

        Parameters
        ----------
        X : Union[pd.DataFrame, pd.Series, np.ndarray, list]
            The Dataframe for the training data

        Returns
        -------
        Union[pd.DataFrame, pd.Series, np.ndarray, list]
            The transformed(numerical values are cast to float32) X data
        """
        # Data type cast could be expensive doing it in a for loop
        # Especially with wide datasets
        # So cast the numerical columns first, without loop
        # Then impute missing values
        if isinstance(X, pd.Series):
            series_name = X.name if X.name else 0
            _X = X.to_frame()
            _X = ONNXTransformer._handle_dtypes_dataframe(_X)[series_name]
        elif isinstance(X, pd.DataFrame):
            _X = ONNXTransformer._handle_dtypes_dataframe(X)
        elif isinstance(X, np.ndarray):
            _X = ONNXTransformer._handle_dtypes_np_array(X)
        else:
            # if users convert pandas dataframe with mixed types to numpy array directly
            # it will turn the whole numpy array into object even though some columns are
            # numerical and some are not. In that case, we need to do extra work to identify
            # which columns are really numerical which for now, we only convert to float32
            # if numpy array is all numerical. else, nothing will be done.
            _X = X
        return _X

    @staticmethod
    def _handle_dtypes_dataframe(X: pd.DataFrame):
        """handle the dtypes for pandas dataframe.

        Parameters
        ----------
        X : pandas.DataFrame
            The Dataframe for the training data

        Returns
        -------
        pandas.DataFrame
            The transformed X data
        """
        dict_astype = {}
        for k, v in zip(X.columns, X.dtypes):
            if "int" in str(v) or "float" in str(v) or "bool" in str(v):
                dict_astype[k] = "float32"
        _X = X.astype(dict_astype)
        if len(dict_astype) > 0:
            logging.warning("Numerical values in `X` are cast to float32.")
        return _X

    @staticmethod
    def _handle_dtypes_np_array(X: np.ndarray):
        """handle the dtypes for pandas dataframe.

        Parameters
        ----------
        X : np.ndarray
            The ndarray for the training data

        Returns
        -------
        np.ndarray
            The transformed X data
        """
        if "int" in str(X.dtype) or "float" in str(X.dtype) or "bool" in str(X.dtype):
            _X = X.astype("float32")
            logging.warning("Numerical values in `X` are cast to float32.")
        else:
            _X = X
        return _X

    def fit(
        self,
        X: Union[pd.DataFrame, pd.Series, np.ndarray, list],
        impute_values: Dict = None,
    ):
        """
        Fits the OnnxTransformer on the dataset
        Parameters
        ----------
        X : Union[pandas.DataFrame, pandas.Series, np.ndarray, list]
            The Dataframe for the training data

        Returns
        -------
        Self: ads.Model
            The fitted estimator
        """
        _X = ONNXTransformer._handle_dtypes(X)
        if isinstance(_X, pd.DataFrame):
            self.dtypes = _X.dtypes
        elif isinstance(_X, np.ndarray):
            self.dtypes = _X.dtype
        self.impute_values = impute_values if impute_values else {}
        self._fitted = True
        return self

    def transform(self, X: Union[pd.DataFrame, pd.Series, np.ndarray, list]):
        """
        Transforms the data for the OnnxTransformer.

        Parameters
        ----------
        X: Union[pandas.DataFrame, pandas.Series, np.ndarray, list]
            The Dataframe for the training data

        Returns
        -------
        Union[pandas.DataFrame, pandas.Series, np.ndarray, list]
            The transformed X data
        """
        assert self._fitted, "Call fit_transform first!"
        if self.dtypes is not None and len(self.dtypes) > 0:
            if isinstance(X, list):
                _X = np.array(X).astype(self.dtypes).tolist()
            else:
                _X = X.astype(self.dtypes)
        else:
            _X = X
        _X = ONNXTransformer._handle_missing_value(_X, impute_values=self.impute_values)
        return _X

    @staticmethod
    def _handle_missing_value(
        X: Union[pd.DataFrame, pd.Series, np.ndarray, list], impute_values: Dict
    ):
        """Impute missing values in X according to impute_values.

        Parameters
        ----------
        X: Union[pandas.DataFrame, pandas.Series, np.ndarray, list]
            The Dataframe for the training data

        Raises
        ------
        Exception if X has only one dim, but imputed_values has multiple values.
        NotImplemented if X has the data type that is not supported.

        Returns
        -------
        Union[pandas.DataFrame, pd.Series, np.ndarray, list]
            The transformed X data
        """
        if isinstance(X, np.ndarray):
            X = ONNXTransformer._handle_missing_value_dataframe(
                pd.DataFrame(X), impute_values=impute_values
            ).values
        elif isinstance(X, list):
            X = ONNXTransformer._handle_missing_value_dataframe(
                pd.DataFrame(X), impute_values=impute_values
            ).values.tolist()
        elif isinstance(X, pd.DataFrame):
            X = ONNXTransformer._handle_missing_value_dataframe(
                X, impute_values=impute_values
            )
        elif isinstance(X, pd.Series):
            X = X.replace(r"^\s*$", np.NaN, regex=True)
            if len(impute_values.keys()) == 1:
                for key, val in impute_values.items():
                    X = X.fillna(val)
            else:
                raise Exception(
                    "Multiple imputed values are provided, but `X` has only one dim."
                )
        else:
            raise NotImplemented(
                f"{type(X)} is not supported. Convert `X` to pandas dataframe or numpy array."
            )
        return X

    @staticmethod
    def _handle_missing_value_dataframe(X: pd.DataFrame, impute_values: Dict):
        for idx, val in impute_values.items():
            if isinstance(idx, int):
                X.iloc[:, idx] = (
                    X.iloc[:, idx].replace(r"^\s*$", np.NaN, regex=True).fillna(val)
                )
            else:
                X.loc[:, idx] = (
                    X.loc[:, idx].replace(r"^\s*$", np.NaN, regex=True).fillna(val)
                )
        return X

    def fit_transform(
        self, X: Union[pd.DataFrame, pd.Series], impute_values: Dict = None
    ):
        """
        Fits, then transforms the data
        Parameters
        ----------
        X: Union[pandas.DataFrame, pandas.Series]
            The Dataframe for the training data

        Returns
        -------
        Union[pandas.DataFrame, pandas.Series]
            The transformed X data
        """
        return self.fit(X, impute_values).transform(X)

    def save(self, filename, **kwargs):
        """
        Saves the Onnx model to disk
        Parameters
        ----------
        filename: Str
            The filename location for where the model should be saved

        Returns
        -------
        filename: Str
            The filename where the model was saved
        """
        export_dict = {
            "impute_values": {
                "value": self.impute_values,
                "dtype": str(type(self.impute_values)),
            },
            "dtypes": {}
            if self.dtypes is None
            else {
                "value": {
                    "index": list(self.dtypes.index),
                    "values": [str(val) for val in self.dtypes.values],
                }
                if isinstance(self.dtypes, pd.Series)
                else str(self.dtypes),
                "dtype": str(type(self.dtypes)),
            },
            "_fitted": {"value": self._fitted, "dtype": str(type(self._fitted))},
        }

        with open(filename, "w") as f:
            json.dump(export_dict, f, sort_keys=True, indent=4, separators=(",", ": "))
        return filename

    @staticmethod
    def load(filename, **kwargs):
        """
        Loads the Onnx model to disk
        Parameters
        ----------
        filename: Str
            The filename location for where the model should be loaded

        Returns
        -------
        onnx_transformer: ONNXTransformer
            The loaded model
        """
        # Make sure you have  pandas, numpy, and sklearn imported
        with open(filename, "r") as f:
            export_dict = json.load(f)

        onnx_transformer = ONNXTransformer()
        for key in export_dict.keys():
            if key not in ["impute_values", "dtypes"]:
                try:
                    setattr(onnx_transformer, key, export_dict[key]["value"])
                except Exception as e:
                    print(
                        f"Warning: Failed to reload {key} from {filename} to OnnxTransformer."
                    )
                    raise e
        if "value" in export_dict["dtypes"]:
            if "index" in export_dict["dtypes"]["value"]:
                onnx_transformer.dtypes = pd.Series(
                    data=[
                        np.dtype(val)
                        for val in export_dict["dtypes"]["value"]["values"]
                    ],
                    index=export_dict["dtypes"]["value"]["index"],
                )
            else:
                onnx_transformer.dtypes = export_dict["dtypes"]["value"]
        else:
            onnx_transformer.dtypes = {}
        onnx_transformer.impute_values = export_dict["impute_values"]["value"]
        return onnx_transformer



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.common.decorator.deprecate import deprecated
from ads.model.extractor.model_info_extractor import ModelInfoExtractor
from ads.model.model_metadata import Framework


class AutoMLExtractor(ModelInfoExtractor):
    """Class that extract model metadata from automl models.

    Attributes
    ----------
    model: object
        The model to extract metadata from.
    estimator: object
        The estimator to extract metadata from.
    """

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self, model):
        self.model = model

    @property
    def framework(self):
        """Extracts the framework of the model.

        Returns
        ----------
        str:
           The framework of the model.
        """
        return Framework.ORACLE_AUTOML

    @property
    def algorithm(self):
        """Extracts the algorithm of the model.

        Returns
        ----------
        object:
           The algorithm of the model.
        """
        return "ensemble"

    @property
    def version(self):
        """Extracts the framework version of the model.

        Returns
        ----------
        str:
           The framework version of the model.
        """
        import automl

        return automl.__version__

    @property
    def hyperparameter(self):
        """Extracts the hyperparameters of the model.

        Returns
        ----------
        dict:
           The hyperparameters of the model.
        """
        return None



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import, division

import pandas as pd
from sklearn.utils.multiclass import type_of_target

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import DiscreteDiscoveryDetector
from ads.type_discovery.typed_feature import (
    OrdinalTypedFeature,
    CategoricalTypedFeature,
)
from ads.common import utils


class DiscreteDetector(DiscreteDiscoveryDetector):

    _max_categorical_values = 100

    def _get_categorical_or_ordinal(self, name, series):
        #
        # categoricals are unordered discreet types
        # ordinals are ordered discreet (int) types
        #

        low_level_type_name = series.dtype.name

        if low_level_type_name == "category" or low_level_type_name == "bool":
            return "categorical"

        else:
            #
            # after removing nulls the new Series might already be categorical
            #
            nulls_removed = pd.Series(list(series.loc[~series.isna()]))
            if (
                nulls_removed.dtype.name == "category"
                or nulls_removed.dtype.name == "bool"
            ):
                return "categorical"

            count_distinct = series.nunique()
            observations = series.size

            tot = type_of_target(
                list(nulls_removed.head(min(nulls_removed.size, 2000)))
            )

            if tot == "binary":
                return "categorical"

            elif tot == "multiclass":
                if count_distinct <= DiscreteDetector._max_categorical_values:
                    if low_level_type_name in utils.numeric_pandas_dtypes():
                        return "ordinal"
                    else:
                        return "categorical"

                if low_level_type_name.startswith(
                    "int"
                ) or low_level_type_name.startswith("float"):
                    if nulls_removed.min() >= 0:
                        if (
                            low_level_type_name.startswith("int")
                            or nulls_removed.sum()
                            == nulls_removed.astype("int64").sum()
                        ):
                            return "ordinal"
                    # by summing all the values and summing all the int values we can know all the values are integers

        return False

    def discover(self, name, series):

        guessed_type = self._get_categorical_or_ordinal(
            name, series.loc[~series.isnull()]
        )

        if guessed_type == "categorical":
            logger.debug("column [{}]/[{}] categorical".format(name, series.dtype))
            return CategoricalTypedFeature.build(name, series)
        elif guessed_type == "ordinal":
            logger.debug("column [{}]/[{}] ordinal".format(name, series.dtype))
            return OrdinalTypedFeature.build(name, series)
        else:
            return False


if __name__ == "__main__":
    dd = DiscreteDetector()

    print(
        dd.discover(
            "str-categorical",
            pd.Series(["a", "a", "a", "b", "c", "a"], dtype="category"),
        )
    )
    print(
        dd.discover(
            "bool-categorical", pd.Series([True, False, True, True, True, None, True])
        )
    )
    print(dd.discover("continuous", pd.Series([None, 3.14, 12.0, 1, 2, 3, None])))
    print(
        dd.discover("int-1-categorical", pd.Series([1, 1, 1, 5, 5, 5, 5, 5, 5, 5, 9]))
    )
    print(dd.discover("int-2-categorical", pd.Series([1, 1, 1, 5, 9])))
    print(dd.discover("real-3-categorical", pd.Series([1.0, 2.0, 3.0, 1.0, 4.0, 5.0])))
    print(
        dd.discover(
            "bool-categorical", pd.Series([True, False, True, True, True, None, True])
        )
    )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Any, Dict

from ads.common.object_storage_details import ObjectStorageDetails
from ads.data_labeling.interface.parser import Parser
from ads.data_labeling.metadata import Metadata


class MetadataParser(Parser):
    """MetadataParser class which parses the metadata from the record."""

    EXPECTED_KEYS = [
        "id",
        "compartmentId",
        "displayName",
        "labelsSet",
        "annotationFormat",
        "datasetSourceDetails",
        "datasetFormatDetails",
    ]

    @staticmethod
    def parse(json_data: Dict[Any, Any]) -> "Metadata":
        """Parses the metadata jsonl file.

        Parameters
        ----------
        json_data: dict
            dictionary format of the metadata jsonl file content.

        Returns
        -------
        Metadata
            Metadata object which contains the useful fields from the metadata jsonl file
        """
        MetadataParser._validate(json_data)
        source_path = ObjectStorageDetails(
            json_data["datasetSourceDetails"]["bucket"],
            json_data["datasetSourceDetails"]["namespace"],
            json_data["datasetSourceDetails"]["prefix"],
        ).path

        records_path = ""
        if "recordFiles" in json_data:
            records_path = ObjectStorageDetails(
                json_data["recordFiles"][0]["bucket"],
                json_data["recordFiles"][0]["namespace"],
                json_data["recordFiles"][0]["path"],
            ).path

        return Metadata(
            source_path=source_path,
            records_path=records_path,
            labels=[clss["name"] for clss in json_data["labelsSet"]],
            dataset_name=json_data["displayName"],
            compartment_id=json_data["compartmentId"],
            dataset_id=json_data["id"],
            annotation_type=json_data["annotationFormat"],
            dataset_type=json_data["datasetFormatDetails"]["formatType"],
        )

    @staticmethod
    def _validate(json_data: Dict[Any, Any]) -> None:
        """Validates the metadata jsonl file to ensure it contains certain fields.

        Parameters
        ----------
        json_data: dict
            dictionary format of the metadata jsonl file content.
        """

        def invalid_message(param):
            return (
                f"The dataset metadata file is invalid. The field '{param}' is required but it is missing. "
                + "Update the metadata file or use the `DataLabeling.export()` method "
                + "to create a new dataset metadata file."
            )

        for k in MetadataParser.EXPECTED_KEYS:
            if k not in json_data:
                raise ValueError(f"{invalid_message(k)}")
        expected_list_format = ["labelsSet", "recordFiles"]
        for k in expected_list_format:
            if k in json_data and not isinstance(json_data[k], list):
                raise ValueError(f"{invalid_message(k)}")
        expected_dict_format = ["datasetSourceDetails", "datasetFormatDetails"]
        for k in expected_dict_format:
            if not isinstance(json_data[k], dict):
                raise ValueError(f"{invalid_message(k)}")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


import click

from ads.common.auth import AuthType
from ads.opctl.utils import suppress_traceback
from ads.opctl.spark.cmds import core_site as core_site_cmd


@click.group("spark")
@click.help_option("--help", "-h")
def commands():
    pass


@commands.command()
@click.option(
    "--auth",
    "-a",
    help="authentication method",
    type=click.Choice(AuthType.values()),
    default=None,
)
@click.option("--oci-config", help="oci config file path", default=None, required=False)
@click.option(
    "--overwrite", "-o", help="overwrite core-site.xml", default=False, is_flag=True
)
@click.option("--oci-profile", help="oci profile to use", default=None, required=False)
@click.option("--debug", "-d", help="set debug mode", is_flag=True, default=False)
@click.help_option("--help", "-h")
def core_site(auth, oci_config, overwrite, oci_profile, debug):
    suppress_traceback(debug)(core_site_cmd)(
        auth,
        oci_config,
        overwrite,
        oci_profile,
    )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


OCI__RUNTIME_TYPE = "OCI__RUNTIME_TYPE"
OCI__RUNTIME_URI = "OCI__RUNTIME_URI"
OCI__RUNTIME_PYTHON_PATH = "OCI__RUNTIME_PYTHON_PATH"
OCI__RUNTIME_GIT_BRANCH = "OCI__RUNTIME_GIT_BRANCH"
OCI__RUNTIME_GIT_COMMIT = "OCI__RUNTIME_GIT_COMMIT"
OCI__RUNTIME_GIT_SECRET_ID = "OCI__RUNTIME_GIT_SECRET_ID"
OCI__CODE_DIR = "OCI__CODE_DIR"


class ClusterConfigToJobSpecConverter:
    def __init__(self, cluster_info):
        self.cluster_info = cluster_info

    def job_def_info(self):
        job = {}
        job["infrastructure"] = self.cluster_info.infrastructure
        cluster = self.cluster_info.cluster
        job["image"] = cluster.image or cluster.main.image or cluster.worker.image
        job["name"] = cluster.name or self.cluster_info.infrastructure.get(
            "displayName"
        )
        job["envVars"] = cluster.config.envVars
        job["envVars"]["OCI__WORK_DIR"] = cluster.work_dir
        job["envVars"]["OCI__EPHEMERAL"] = cluster.ephemeral
        job["envVars"]["OCI__CLUSTER_TYPE"] = cluster.type.upper()
        job["envVars"]["OCI__WORKER_COUNT"] = (
            cluster.worker.replicas if cluster.worker is not None else 0
        )
        if cluster.ps is not None:
            job["envVars"]["OCI__PS_COUNT"] = cluster.ps.replicas

        job["envVars"]["OCI__START_ARGS"] = cluster.config.cmd_args.strip()
        job["envVars"]["OCI__ENTRY_SCRIPT"] = self.cluster_info.runtime.entry_point

        if self.cluster_info.runtime.type:
            optional_runtime_fields = [
                {"attr": "type", "env": OCI__RUNTIME_TYPE},
                {"attr": "uri", "env": OCI__RUNTIME_URI},
                {"attr": "branch", "env": OCI__RUNTIME_GIT_BRANCH},
                {"attr": "commit", "env": OCI__RUNTIME_GIT_COMMIT},
                {"attr": "git_secret_id", "env": OCI__RUNTIME_GIT_SECRET_ID},
                {"attr": "code_dir", "env": OCI__CODE_DIR},
                {"attr": "python_path", "env": OCI__RUNTIME_PYTHON_PATH},
            ]
            for field in optional_runtime_fields:
                val = getattr(self.cluster_info.runtime, field["attr"], None)
                if val:
                    job["envVars"][field["env"]] = val

        runtime_args = self.cluster_info.runtime.args
        if isinstance(self.cluster_info.runtime.args, list):
            runtime_args = " ".join([str(v) for v in self.cluster_info.runtime.args])
        if runtime_args:
            job["envVars"]["OCI__ENTRY_SCRIPT_ARGS"] = runtime_args
        if self.cluster_info.runtime.kwargs:
            job["envVars"][
                "OCI__ENTRY_SCRIPT_KWARGS"
            ] = self.cluster_info.runtime.kwargs
        job["envVars"].update(self.cluster_info.runtime.envVars)
        job["envVars"] = {k: str(job["envVars"][k]) for k in job["envVars"]}

        if self.cluster_info.cluster.certificate:
            job["envVars"][
                "OCI__CERTIFICATE_OCID"
            ] = self.cluster_info.cluster.certificate.cert_ocid
            job["envVars"][
                "OCI__CERTIFICATE_KEY_DOWNLOAD_LOCATION"
            ] = self.cluster_info.cluster.certificate.key_download_location
            job["envVars"][
                "OCI__CERTIFICATE_DOWNLOAD_LOCATION"
            ] = self.cluster_info.cluster.certificate.cert_download_location
            job["envVars"][
                "OCI__CERTIFICATE_AUTHORITY_OCID"
            ] = self.cluster_info.cluster.certificate.ca_ocid
            job["envVars"][
                "OCI__CA_DOWNLOAD_LOCATION"
            ] = self.cluster_info.cluster.certificate.ca_download_location

        return job

    def job_run_info(self, jobType):
        jobrun = {}
        jobTypeConfig = getattr(self.cluster_info.cluster, jobType)
        if jobTypeConfig is not None:
            jobrun["name"] = jobTypeConfig.name or jobType
            jobrun["envVars"] = jobTypeConfig.config.envVars
            if jobTypeConfig.config.cmd_args:
                jobrun["envVars"][
                    "OCI__START_ARGS"
                ] = jobTypeConfig.config.cmd_args.strip()

            jobrun["envVars"]["OCI__MODE"] = jobType.upper()
            jobrun["envVars"] = {
                k: str(jobrun["envVars"][k]) for k in jobrun["envVars"]
            }
        return jobrun



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import, division

import re

import pandas as pd
from pandas.api.types import is_string_dtype

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import PhoneNumberTypedFeature


class PhoneNumberDetector(AbstractTypeDiscoveryDetector):

    _pattern_string = r"^(\+?\d{1,2}[\s-])?\(?(\d{3})\)?[\s.-]?\d{3}[\s.-]?\d{4}$"

    def __init__(self):
        self.cc = re.compile(PhoneNumberDetector._pattern_string, re.VERBOSE)

    def is_phone_number(self, name, values):
        cc = re.compile(PhoneNumberDetector._pattern_string, re.VERBOSE)
        return all([self.cc.match(str(x)) for x in values])

    def discover(self, name, series):
        if is_string_dtype(series):
            candidates = series.loc[~series.isnull()]

            if self.is_phone_number(name, candidates.head(1000)):
                logger.debug("column [{}]/[{}] phone number".format(name, series.dtype))
                return PhoneNumberTypedFeature.build(name, series)

        return False


if __name__ == "__main__":
    dd = PhoneNumberDetector()

    # test the postive case:
    test_series_1 = [
        "408-456-7890",
        "(123) 456-7890",
        "650 456 7890",
        "650.456.7890",
        "+91 (123) 456-7890",
    ]

    print(dd.discover("test_series_1", pd.Series(test_series_1)))

    # test the negative case
    test_series_2 = ["1234567890", "1234567890", "123456890", "1234567890"]

    print(dd.discover("test_series_2", pd.Series(test_series_2)))

    test_series_3 = ["1-640-124-5367", "1-573-916-4412"]

    print(dd.discover("test_series_3", pd.Series(test_series_3)))



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from numbers import Integral, Number
from time import time

import numpy as np

import sklearn
from sklearn.base import BaseEstimator, clone, is_classifier
from sklearn.model_selection import BaseCrossValidator  # NOQA
from sklearn.model_selection import check_cv, cross_validate
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.utils import check_random_state
from sklearn.utils.metaestimators import _safe_split

try:
    from sklearn.utils import _safe_indexing as sklearn_safe_indexing
except:
    from sklearn.utils import safe_indexing as sklearn_safe_indexing

try:
    from sklearn.metrics import check_scoring
except:
    from sklearn.metrics.scorer import check_scoring
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class _Objective(object):
    """
    Callable that implements objective function.


    Parameters
    ----------
        model:
            Object to use to fit the data. This is assumed to implement the
            scikit-learn estimator interface. Either this needs to provide
            ``score``, or ``scoring`` must be passed.
        X:
            Training data.
        y:
            Target variable.
        cv:
            Cross-validation strategy.
        enable_pruning:
            If :obj:`True`, pruning is performed in the case where the
            underlying estimator supports ``partial_fit``.
        error_score:
            Value to assign to the score if an error occurs in fitting. If
            'raise', the error is raised. If numeric,
            ``sklearn.exceptions.FitFailedWarning`` is raised. This does not
            affect the refit step, which will always raise the error.
        fit_params:
            Parameters passed to ``fit`` one the estimator.
        groups:
            Group labels for the samples used while splitting the dataset into
            train/validation set.
        max_iter,
            max number of iteration for ``partial_fit``
        return_train_score:
            If :obj:`True`, training scores will be included. Computing
            training scores is used to get insights on how different
            hyperparameter settings impact the overfitting/underfitting
            trade-off. However computing training scores can be
            computationally expensive and is not strictly required to select
            the hyperparameters that yield the best generalization
            performance. Hence, we default it to be False and do not expose
            this parameter to the users.
        scoring:
            Scorer function.
        scoring_name:
            name of the Scorer function.
        step_name:
            step name of the estimator in a pipeline

        Returns
        -------
            the objective score

    """

    def __init__(
        self,
        model,  # type: Union[BaseEstimator, Pipeline]
        param_distributions,  # type: Mapping[str, distributions.BaseDistribution]
        cv,  # type: BaseCrossValidator
        enable_pruning,  # type: bool
        error_score,  # type: Union[Number, str]
        fit_params,  # type: Dict[str, Any]
        groups,  # type: Optional[OneDimArrayLikeType]
        max_iter,  # type: int
        return_train_score,  # type: bool
        scoring,  # type: Callable[..., Number]
        scoring_name,  # type: str
        step_name,  # type: str
    ):
        # type: (...) -> None
        self.model = model
        self.param_distributions = param_distributions
        self.cv = cv
        self.enable_pruning = enable_pruning
        self.error_score = error_score
        self.fit_params = fit_params
        self.groups = groups
        self.max_iter = max_iter
        self.return_train_score = return_train_score
        self.scoring = scoring
        self.scoring_name = scoring_name
        self.step_name = step_name

    def __call__(self, X, y, trial):
        # type: (trial_module.Trial) -> float

        estimator = clone(self.model)
        params = self._get_params(trial, self.param_distributions)
        params = self._extract_max_iter(params)
        estimator.set_params(**params)

        if self.enable_pruning:
            scores = self._cross_validate_with_pruning(X, y, trial, estimator)
        else:
            scores = cross_validate(
                estimator,
                X,
                y,
                cv=self.cv,
                error_score=self.error_score,
                fit_params=self.fit_params,
                groups=self.groups,
                return_train_score=self.return_train_score,
                scoring=self.scoring,
            )

        self._store_scores(trial, scores, self.scoring_name)

        return trial.user_attrs["mean_test_score"]

    def _extract_max_iter(self, params):
        if self.enable_pruning:
            max_iter_name = "max_iter"
            if self.step_name:
                max_iter_name = self.step_name + "__" + max_iter_name
            if max_iter_name in params:
                self.max_iter = params.pop(max_iter_name)
        return params

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def _cross_validate_with_pruning(
        self,
        X,
        y,
        trial,  # type: trial_module.Trial
        estimator,  # type: BaseEstimator
    ):
        # type: (...) -> Dict[str, OneDimArrayLikeType]

        if is_classifier(estimator):
            partial_fit_params = self.fit_params.copy()
            classes = np.unique(y)

            partial_fit_params.setdefault("classes", classes)

        else:
            partial_fit_params = self.fit_params.copy()

        n_splits = self.cv.get_n_splits(X, y, groups=self.groups)

        estimators = [clone(estimator) for _ in range(n_splits)]
        scores = {
            "fit_time": np.zeros(n_splits),
            "score_time": np.zeros(n_splits),
            "test_score": np.empty(n_splits),
        }

        if self.return_train_score:
            scores["train_score"] = np.empty(n_splits)

        for step in range(self.max_iter):
            for i, (train, test) in enumerate(self.cv.split(X, y, groups=self.groups)):

                out = self._partial_fit_and_score(
                    X, y, estimators[i], train, test, partial_fit_params
                )

                if self.return_train_score:
                    scores["train_score"][i] = out.pop(0)

                scores["test_score"][i] = out[0]
                scores["fit_time"][i] += out[1]
                scores["score_time"][i] += out[2]

            intermediate_value = np.nanmean(scores["test_score"])

            trial.report(intermediate_value, step=step)

            if trial.should_prune():

                self._store_scores(trial, scores, self.scoring_name)

                raise optuna.TrialPruned(f"trial was pruned at iteration {step}.")
        return scores

    def _get_params(self, trial, param_distributions):
        # type: (trial_module.Trial) -> Dict[str, Any]

        return {
            name: trial._suggest(name, distribution.get_distribution())
            for name, distribution in param_distributions.items()
        }

    def _partial_fit_and_score(
        self,
        X,
        y,
        estimator,  # type: BaseEstimator
        train,  # type: List[int]
        test,  # type: List[int]
        partial_fit_params,  # type: Dict[str, Any]
    ):
        # type: (...) -> List[Number]

        X_train, y_train = _safe_split(estimator, X, y, train)
        X_test, y_test = _safe_split(estimator, X, y, test, train_indices=train)

        start_time = time()

        try:
            estimator.partial_fit(X_train, y_train, **partial_fit_params)

        except Exception as e:
            if self.error_score == "raise":
                raise e

            elif isinstance(self.error_score, Number):
                fit_time = time() - start_time
                test_score = self.error_score
                score_time = 0.0

                if self.return_train_score:
                    train_score = self.error_score

            else:
                raise ValueError("error_score must be 'raise' or numeric.")

        else:
            fit_time = time() - start_time
            # Required for type checking but is never expected to fail.
            assert isinstance(fit_time, Number)
            scoring_start_time = time()
            test_score = self.scoring(estimator, X_test, y_test)
            score_time = time() - scoring_start_time
            # Required for type checking but is never expected to fail.
            assert isinstance(score_time, Number)

            if self.return_train_score:
                train_score = self.scoring(estimator, X_train, y_train)

        ret = [test_score, fit_time, score_time]

        if self.return_train_score:
            ret.insert(0, train_score)

        return ret

    def _store_scores(self, trial, scores, scoring_name):
        # type: (trial_module.Trial, Dict[str, OneDimArrayLikeType]) -> None

        trial.set_user_attr("metric", scoring_name)

        for name, array in scores.items():
            if name in ["test_score", "train_score"]:
                for i, score in enumerate(array):
                    trial.set_user_attr("split{}_{}".format(i, name), score)
            trial.set_user_attr("mean_{}".format(name), np.nanmean(array))
            trial.set_user_attr("std_{}".format(name), np.nanstd(array))



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a ZipCode feature type.

Classes:
    ZipCode
        The ZipCode feature type.

Functions:
    default_handler(data: pd.Series) -> pd.Series
        Processes given data and indicates if the data matches requirements.
"""
import matplotlib.pyplot as plt
import pandas as pd
import re
from ads.feature_engineering.feature_type.string import String
from ads.feature_engineering.utils import (
    _count_unique_missing,
    _to_lat_long,
    _plot_gis_scatter,
    _zip_code,
)
from ads.feature_engineering import schema

PATTERN = re.compile(r"^[0-9]{5}(?:-[0-9]{4})?$", re.VERBOSE)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: pd.Series
        The data to process.

    Returns
    -------
    pd.Series: The logical list indicating if the data matches requirements.
    """

    def _is_zip_code(x: any):
        return (
            not pd.isnull(x) and isinstance(x, str) and re.match(PATTERN, x) is not None
        )

    return data.apply(lambda x: True if _is_zip_code(x) else False)


class ZipCode(String):
    """Type representing postal code.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows the geometry distribution base on location of zipcode.


    Example
    -------
    >>> from ads.feature_engineering.feature_type.zip_code import ZipCode
    >>> import pandas as pd
    >>> import numpy as np
    >>> s = pd.Series(["94065", "90210", np.NaN, None], name='zipcode')
    >>> ZipCode.validator.is_zip_code(s)
    0     True
    1     True
    2    False
    3    False
    Name: zipcode, dtype: bool
    """

    description = "Type representing postal code."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count) and missing(count).

        Examples
        --------
        >>> zipcode = pd.Series([94065, 90210, np.NaN, None], name='zipcode')
        >>> zipcode.ads.feature_type = ['zip_code']
        >>> zipcode.ads.feature_stat()
            Metric  Value
        0	count	4
        1	unique	2
        2	missing	2

        Returns
        -------
        Pandas Dataframe
            Summary statistics of the Series provided.
        """
        return _count_unique_missing(x)

    @staticmethod
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows the geometry distribution base on location of zipcode.

        Examples
        --------
        >>> zipcode = pd.Series([94065, 90210, np.NaN, None], name='zipcode')
        >>> zipcode.ads.feature_type = ['zip_code']
        >>> zipcode.ads.feature_plot()
        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the ZipCode feature type.
        """
        gis = _to_lat_long(x.loc[default_handler(x)], _zip_code())
        if len(gis.index):
            return _plot_gis_scatter(gis, "longitude", "latitude")

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> zipcode = pd.Series([94065, 90210, np.NaN, None], name='zipcode')
        >>> zipcode.ads.feature_type = ['zip_code']
        >>> zipcode.ads.feature_domain()
        constraints: []
        stats:
            count: 4
            missing: 2
            unique: 2
        values: ZipCode

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the ZipCode feature type.
        """

        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [],
        )


ZipCode.validator.register("is_zip_code", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import absolute_import, print_function

import collections
import contextlib
import copy
import fnmatch
import glob
import json
import math
import os
import random
import re
import shutil
import string
import sys
import tempfile
from datetime import datetime
from enum import Enum
from io import DEFAULT_BUFFER_SIZE
from pathlib import Path
from textwrap import fill
from typing import Dict, Optional, Union
from urllib import request
from urllib.parse import urlparse

import fsspec
import matplotlib as mpl
import numpy as np
import pandas as pd
from ads.common import logger
from ads.common.decorator.deprecate import deprecated
from ads.common.word_lists import adjectives, animals
from ads.dataset.progress import TqdmProgressBar
from cycler import cycler
from pandas.core.dtypes.common import is_datetime64_dtype, is_numeric_dtype
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from ads.common import logger
from ads.common.decorator.deprecate import deprecated
from ads.common.decorator.runtime_dependency import (
    OptionalDependency,
    runtime_dependency,
)
from ads.common.word_lists import adjectives, animals
from ads import config
from ads.dataset.progress import DummyProgressBar, TqdmProgressBar

from . import auth as authutil

# For Model / Model Artifact libraries
lib_translator = {"sklearn": "scikit-learn"}
module_ignore = ["builtins", "ads", "automl", "mlx"]

# up-sample if length of dataframe is less than or equal to MAX_LEN_FOR_UP_SAMPLING
MAX_LEN_FOR_UP_SAMPLING = 5000

# down-sample if ratio of minority to majority class is greater than or equal to MIN_RATIO_FOR_DOWN_SAMPLING
MIN_RATIO_FOR_DOWN_SAMPLING = 1 / 20

# Maximum distinct values by cardinality will be used for plotting
MAX_DISPLAY_VALUES = 10

# par link of the index json file.
PAR_LINK = "https://objectstorage.us-ashburn-1.oraclecloud.com/p/Ri7zFc_h91sxMdgnza9Qnqw3Ina8hf8wzDvEpAnUXMDOnUR1U1fpsaBUjUfgPgIq/n/ociodscdev/b/service-conda-packs/o/service_pack/index.json"

random_state = 42
test_size = 0.3
date_format = "%Y-%m-%d %H:%M:%S"

# at this time, we only support regression and classification tasks.
ml_task_types = Enum(
    "ml_task_types",
    "REGRESSION BINARY_CLASSIFICATION MULTI_CLASS_CLASSIFICATION BINARY_TEXT_CLASSIFICATION "
    "MULTI_CLASS_TEXT_CLASSIFICATION UNSUPPORTED",
)

mpl.rcParams["image.cmap"] = "BuGn"
mpl.rcParams["axes.prop_cycle"] = cycler(
    color=["teal", "blueviolet", "forestgreen", "peru", "y", "dodgerblue", "r"]
)

# sqlalchemy engines
_engines = {}

ORACLE_DEFAULT_PORT = 1521
MYSQL_DEFAULT_PORT = "3306"

# Maximum number of columns of data to extract model schema.
DATA_SCHEMA_MAX_COL_NUM = 2000

# dimention of np array which can be converted to pd dataframe
DIMENSION = 2

# declare custom exception class


class FileOverwriteError(Exception):
    pass


def get_cpu_count():
    """
    Returns the number of CPUs available on this machine
    """
    return os.cpu_count()


@deprecated(
    "2.5.2", details="This method is being deprecated in favor of `get_cpu_count`"
)
def get_compute_accelerator_ncores():
    return get_cpu_count()


@deprecated(
    "2.5.10",
    details="Deprecated, use: from ads.common.auth import AuthState;"
    "oci_config_location=AuthState().oci_config_path; profile=AuthState().oci_key_profile",
)
def get_oci_config():
    """
    Returns the OCI config location, and the OCI config profile.
    """
    oci_config_location = os.environ.get(
        "OCI_CONFIG_LOCATION", f"{os.environ['HOME']}/.oci/config"
    )  # os.environ['HOME'] == home/datascience
    oci_config_profile = os.environ.get("OCI_CONFIG_PROFILE", "DEFAULT")
    return oci_config_location, oci_config_profile


@deprecated(
    "2.5.10",
    details="Deprecated, use: from ads.common.auth import AuthState; os.path.dirname(AuthState().oci_config_path)",
)
def oci_key_location():
    """
    Returns the OCI key location
    """
    return os.environ.get(
        "OCI_CONFIG_DIR", os.path.join(os.path.expanduser("~"), ".oci")
    )


@deprecated(
    "2.5.10",
    details="Deprecated, use: from ads.common.auth import AuthState; AuthState().oci_config_path",
)
def oci_config_file():
    """
    Returns the OCI config file location
    """
    return os.path.join(oci_key_location(), "config")


@deprecated(
    "2.5.10",
    details="Deprecated, use: from ads.common.auth import AuthState; AuthState().oci_key_profile",
)
def oci_config_profile():
    """
    Returns the OCI config profile location.
    """
    return os.environ.get("OCI_CONFIG_PROFILE", "DEFAULT")


def numeric_pandas_dtypes():
    """
    Returns a list of the "numeric" pandas data types
    """
    return ["int16", "int32", "int64", "float16", "float32", "float64"]


@deprecated(
    "2.5.10",
    details="Deprecated, use: ads.set_auth(auth='api_key', oci_config_location='~/.oci/config', profile='DEFAULT')",
)
def set_oci_config(oci_config_location, oci_config_profile):
    """
    :param oci_config_location: location of the config file, for example, ~/.oci/config
    :param oci_config_profile: The profile to load from the config file.  Defaults to "DEFAULT"
    """
    if not os.path.exists(f"{oci_config_location}"):
        raise ValueError("The oci config file doesn't exist.")
    os.environ["OCI_CONFIG_LOCATION"] = oci_config_location
    os.environ["OCI_CONFIG_PROFILE"] = oci_config_profile


def random_valid_ocid(prefix="ocid1.dataflowapplication.oc1.iad"):
    """
    Generates a random valid ocid.

    Parameters
    ----------
    prefix: `str`
        A prefix, corresponding to a region location.

    Returns
    -------
    ocid: `str`
        a valid ocid with the given prefix.
    """
    left, right = prefix.rsplit(".", 1)
    fake = "".join([random.choice(string.ascii_lowercase) for i in range(60)])
    return f"{left}.{fake}"


def get_dataframe_styles(max_width=75):
    """Styles used for dataframe, example usage:

        df.style\
            .set_table_styles(utils.get_dataframe_styles())\
            .set_table_attributes('class=table')\
            .render())

    Returns
    -------
    styles: array
        A list of dataframe table styler styles.
    """

    alt_props = [
        ("background-color", "#F8F8F8"),
    ]

    th_props = [
        ("font-size", "12px"),
        ("text-align", "center"),
        ("font-weight", "bold"),
        ("background-color", "#D3D3D3"),
        ("padding-left", "5px"),
        ("padding-right", "5px"),
        ("text-align", "right"),
    ]

    td_props = [
        ("font-size", "12px"),
        ("text_wrap", False),
        ("white-space", "nowrap"),
        ("overflow", "hidden"),
        ("text-overflow", "ellipsis"),
        ("max-width", f"{max_width}px"),
    ]

    hover_props = [("background-color", "#D9EDFD")]

    styles = [
        dict(selector="tbody tr:nth-child(even)", props=alt_props),
        dict(selector="tbody tr:hover", props=hover_props),
        dict(selector="th", props=th_props),
        dict(selector="td", props=td_props),
    ]

    return styles


def get_bootstrap_styles():
    """
    Returns HTML bootstrap style information
    """
    return """<style>

        code {
            padding: 2px 4px;
            font-size: 90%;
            color: #c7254e;
            background-color: #f9f2f4;
            border-radius: 4px;
            font-family: Menlo,Monaco,Consolas,"Courier New",monospace;
        }

        .label {
            display: inline;
            padding: .2em .6em .3em;
            font-weight: 700;
            line-height: 1;
            color: #fff;
            font-size: 85%;
            text-align: center;
            white-space: nowrap;
            vertical-align: baseline;
            border-radius: .25em;
        }

        .label-high-cardinality {
            background-color: #fe7c1b;
        }

        .label-missing {
            background-color: #214761;
        }

        .label-zeros {
            background-color: #333796;
        }

        .label-warning {
            background-color: #e2007e;
        }

        .label-skew {
            background-color: #ffdb58;
            color: black;
        }

        .label-duplicate-rows {
            background-color: #d90773;
        }
    </style>"""


def highlight_text(text):
    """Returns text with html highlights.
    Parameters
    ----------
    text: String
      The text to be highlighted.

    Returns
    -------
    ht: String
        The text with html highlight information.
    """
    return f"""<code style="background:yellow; color:black; padding-top: 5px; padding-bottom: 5px">
        {text}
    </code>""".strip()


def horizontal_scrollable_div(html):
    """Wrap html with the necessary html to make horizontal scrolling possible.

    Examples
    ________

    display(HTML(utils.horizontal_scrollable_div(my_html)))

    Parameters
    ----------
    html: str
        Your HTML to wrap.

    Returns
    -------
    type
        Wrapped HTML.
    """

    return f"""
        <style>
            .mostly-customized-scrollbar {{
              display: block;
              width: 100%;
              overflow: auto;
            }}

            .mostly-customized-scrollbar::-webkit-scrollbar {{
              width: 5px;
              height: 8px;
              background-color: #aaa;
            }}

            .mostly-customized-scrollbar::-webkit-scrollbar-thumb {{
                background: #000;
                border-radius: 10px;
            }}
        </style>

        <div style="width=100%; display: flex; flex-wrap: nowrap; overflow-x: auto;">
            <div class="mostly-customized-scrollbar">
                {html}
            </div>
        </div>
    """


def is_notebook():
    """Returns true if the environment is a jupyter notebook."""
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":  # pragma: no cover
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except ModuleNotFoundError or NameError:
        return False  # Probably standard Python interpreter


def is_test():  # pragma: no cover
    """
    Returns true if ADS is in test mode.
    """
    from ads import test_mode

    return test_mode


@deprecated(
    "2.6.8",
    details="Deprecated, use: ads.set_auth(auth='resource_principal')",
)
def is_resource_principal_mode():  # pragma: no cover
    """
    Returns true if ADS is in resource principal mode.
    """
    from ads import resource_principal_mode

    return resource_principal_mode


@deprecated(
    "2.6.8",
    details="Deprecated, use: from ads.common.auth import AuthState; AuthState().oci_config_path",
)
def oci_config_location():  # pragma: no cover
    """
    Returns oci configuration file location.
    """
    from ads.common.auth import AuthState

    return AuthState().oci_config_path


@deprecated(
    "2.6.8",
    details="Deprecated, use: from ads.common.auth import AuthState; AuthState().oci_key_profile",
)
def oci_key_profile():  # pragma: no cover
    """
    Returns key profile value specified in oci configuration file.
    """
    from ads.common.auth import AuthState

    return AuthState().oci_key_profile


def is_documentation_mode():  # pragma: no cover
    """
    Returns true if ADS is in documentation mode.
    """
    from ads import documentation_mode

    return documentation_mode


def is_debug_mode():  # pragma: no cover
    """
    Returns true if ADS is in debug mode.
    """
    from ads import debug_mode

    return debug_mode


@deprecated("2.3.1")
@runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
def print_user_message(
    msg, display_type="tip", see_also_links=None, title="Tip"
):  # pragma: no cover
    """This method is deprecated and will be removed in future releases.
    Prints in html formatted block one of tip|info|warn type.

    Parameters
    ----------
    msg : str or list
        The actual message to display.
        display_type is "module', msg can be a list of [module name, module package name], i.e. ["automl", "ads[ml]"]
    display_type : str (default 'tip')
        The type of user message.
    see_also_links : list of tuples in the form of [('display_name', 'url')]
    title : str (default 'tip')
        The title of user message.
    """

    if display_type.lower() == "error" and not is_documentation_mode():
        print("ERROR: {}".format(re.sub("<[^>]*>", "", msg)))

    if display_type.lower() == "module":
        if isinstance(msg, list()):
            module_name = msg[0]
            module_pkg = msg[1]
        else:
            module_name, module_pkg = msg, msg
        print(
            f"ERROR: {module_name} module not found. Make sure you have installed {module_pkg} in order to download all of necessary modules."
        )

    if is_documentation_mode() and is_notebook():

        if display_type.lower() == "tip":

            if "\n" in msg:
                t = "<b>{}:</b>".format(title.upper().strip()) if title else ""

                user_message = "{}{}".format(
                    t,
                    "".join(
                        [
                            "<br>&nbsp;&nbsp;+&nbsp;{}".format(x.strip())
                            for x in msg.strip().split("\n")
                        ]
                    ),
                )
            else:
                user_message = "{}".format(msg.strip().replace("\n", "<br>"))

            from IPython.core.display import HTML, display

            display(
                HTML(
                    f"""
                <div style="padding: 7px;
                            border-radius: 4px;
                            background-color: #d4ecd9;
                            margin_bottom: 5px;">
                    <p>{user_message}</p>
                </div>

                """
                )
            )

        elif display_type.lower() == "warning":
            user_message = "{}".format(msg.strip().replace("\n", "<br>"))

            display(
                HTML(
                    f"""
                <div style="padding: 7px;
                            border-radius: 4px;
                            background-color: #fcc5c5;
                            margin_bottom: 5px;">
                    <h3>Warning:</h3>
                    <p>{user_message}</p>
                </div>

                """
                )
            )

        elif display_type.lower() == "error":
            user_message = "{}".format(msg.strip().replace("\n", "<br>"))

            display(
                HTML(
                    f"""
                <div style="padding: 7px;
                            border-radius: 4px;
                            background-color: #4f053b;
                            color: white;
                            margin_bottom: 5px;">
                    <h2>Error:</h2>
                    <p>{user_message}</p>
                </div>

                """
                )
            )

        elif display_type.startswith("info"):

            user_message = msg.strip().replace("\n", "<br>")

            if see_also_links:
                see_also_html = f"""
                <ul>
                {'<li>'.join([f"<a src='{url}'>{display_name}</a>"
                              for (display_name, url) in see_also_links])}
                </ul>
                """
            else:
                see_also_html = ""

            if title:
                title_html = f"""<div style="padding: 5px;
                            color: #588864;
                            border_bottom: 1px solid grey;
                            margin_bottom: 5px;">
                    <h2>{title.upper()}</h2>
                </div>"""
            else:
                title_html = ""

            display(
                HTML(
                    f"""
            <br>

            <div style="width: calc(100% -20px);
                        border-left: 8px solid #588864;
                        margin: 10px, 0, 10px, 0px;
                        padding: 10px">

                {title_html}
                <p>{user_message}</p>
                {see_also_html}
            </div>

            """
                )
            )


# take a series which can be interpreted as a dict, index=key, this
# function sorts by the values and takes the top-n values, returning
# a new series
#
def truncate_series_top_n(series, n=24):
    """
    take a series which can be interpreted as a dict, index=key, this
    function sorts by the values and takes the top-n values, and returns
    a new series
    """
    return series.sort_values(ascending=False).head(n)


#
# take a sequence (<string>, list(<string>), tuple(<string>), pd.Series(<string>) and Ellipsis'ize them at position n
#
def ellipsis_strings(raw, n=24):
    """
    takes a sequence (<string>, list(<string>), tuple(<string>), pd.Series(<string>) and Ellipsis'ize them at position n
    """
    if isinstance(raw, pd.core.indexes.base.Index):
        sequence = raw.astype(str).to_list()

    if isinstance(raw, str):
        return ellipsis_strings([raw], n)[0]

    sequence = list(raw) if not isinstance(raw, list) else raw

    result = []
    for s in sequence:

        if len(str(s)) <= n:
            result.append(s)
        else:
            n2 = int(n) // 2 - 3
            n1 = n - n2 - 3
            result.append("{0}...{1}".format(s[:n1], s[-n2:]))

    return result


def first_not_none(itr):
    """
    Returns the first non-none result from an iterable,
    similar to any() but return value not true/false
    """
    for x in itr:
        if x:
            return x
    return None


#
# checks to see if object is the same class as cls
#
def is_same_class(obj, cls):
    """
    checks to see if object is the same class as cls
    """
    if isinstance(cls, (list, tuple)):
        return any([obj.__class__.__name__ == x.__name__ for x in cls])
    else:
        return obj.__class__.__name__ == cls.__name__


def replace_spaces(lst):
    """
    Replace all spaces with underscores for strings in the list.

    Requires that the list contains strings for each element.

    lst: list of strings
    """
    return [s.replace(" ", "_") for s in lst]


def get_progress_bar(
    max_progress: int, description: str = "Initializing", verbose: bool = False
) -> TqdmProgressBar:
    """Returns an instance of the TqdmProgressBar class.

    Parameters
    ----------
    max_progress: int
        The number of steps for the progressbar.
    description: (str, optional). Defaults to "Initializing".
        The first step description.
    verbose: (bool, optional). Defaults to `False`
        If the progress should show the debug information.

    Returns
    -------
    TqdmProgressBar
        An instance of the TqdmProgressBar.
    """
    return TqdmProgressBar(
        max_progress, description=description, verbose=verbose or is_debug_mode()
    )


class JsonConverter(json.JSONEncoder):
    def default(self, obj):
        """
        Converts an object to JSON based on its type

        Parameters
        ----------
        obj: Object
            An object which is being converted to Json, supported types are pandas Timestamp, series, dataframe, or categorical or numpy ndarrays.

        Returns
        -------
        Json: A json repersentation of the object.
        """
        if isinstance(obj, pd.Timestamp):
            return obj.__str__()
        if isinstance(obj, pd.Series):
            return obj.values
        if isinstance(obj, pd.Categorical):
            return obj.get_values()
        if isinstance(obj, pd.DataFrame):
            return json.loads(obj.to_json())
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):
            return int(obj)
        elif isinstance(
            obj, (np.float_, np.float16, np.float32, np.float64, np.double)
        ):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def split_data(X, y, random_state=random_state, test_size=test_size):
    """
    Splits data using Sklearn based on the input type of the data.

    Parameters
    ----------
    X: a Pandas Dataframe
        The data points.
    y: a Pandas Dataframe
        The labels.
    random_state: int
        A random state for reproducability.
    test_size: int
        The number of elements that should be included in the test dataset.
    """
    return train_test_split(
        X, y, train_size=1 - test_size, test_size=test_size, random_state=random_state
    )


@runtime_dependency(module="sqlalchemy", install_from=OptionalDependency.DATA)
def get_sqlalchemy_engine(connection_url, *args, **kwargs):
    """
    The SqlAlchemny docs say to use a single engine per connection_url, this class will take
    care of that.

    Parameters
    ----------

    connection_url: string
        The URL to connect to

    Returns
    -------
    engine: SqlAlchemny engine
        The engine from which SqlAlchemny commands can be ran on
    """
    global _engines
    if connection_url not in _engines:
        #
        # Note: pool_recycle=1 is used here because sqlalchemy is free to drop inactive
        # connections. This will make sure they are recycled and available when we
        # need them.
        #
        # DAR: note: use echo=True to log engine output
        _engines[connection_url] = sqlalchemy.create_engine(
            connection_url, pool_recycle=10, *args, **kwargs
        )

    return _engines[connection_url]


def inject_and_copy_kwargs(kwargs, **args):
    """Takes in a dictionary and returns a copy with the args injected

    Examples
    ________
    >>> foo(arg1, args, utils.inject_and_copy_kwargs(kwargs, arg3=12, arg4=42))

    Parameters
    ----------
    kwargs : dict
        The original `kwargs`.
    **args : type
        A series of arguments, foo=42, bar=12 etc

    Returns
    -------
    d: dict
        new dictionary object that you can use in place of kwargs

    """

    d = kwargs.copy()
    for k, v in args.items():
        if k not in kwargs:
            d[k] = v  # inject args iff not already found
    return d


def flatten(d, parent_key=""):
    """
    Flattens nested dictionaries to a single layer dictionary

    Parameters
    ----------
    d : dict
        The dictionary that needs to be flattened
    parent_key : str
        Keys in the dictionary that are nested

    Returns
    -------
    a_dict: dict
        a single layer dictionary
    """
    items = []
    for k, v in d.items():
        new_key = k if parent_key else k
        if isinstance(v, collections.abc.MutableMapping):
            items.extend(flatten(v, new_key).items())
        else:
            items.append((new_key, v))

    return dict(items)


def wrap_lines(li, heading=""):
    """
    Wraps the elements of iterable into multi line string of fixed width
    """
    return heading + "\n" + fill(str(list(li)), width=30) if len(li) > 0 else ""


def get_base_modules(model):
    """
    Get the base modules from an ADS model
    """
    add_bases = []
    if hasattr(model, "est"):
        add_bases = get_base_modules(model.est)
    try:
        if hasattr(model, "steps") and isinstance(model.steps, list):
            [add_bases.extend(get_base_modules(step)) for _, step in model.steps]
    except:
        pass
    return (
        add_bases + list(type(model).__mro__)
        if hasattr(type(model), "__mro__")
        else add_bases
    )


def extract_lib_dependencies_from_model(model) -> dict:
    """
    Extract a dictionary of library dependencies for a model

    Parameters
    ----------
    model

    Returns
    -------
    Dict: A dictionary of library dependencies.
    """
    from pkg_resources import get_distribution

    module_versions = {}
    modules_to_include = set(
        mod.__module__.split(".")[0]
        for mod in get_base_modules(model)
        if hasattr(mod, "__module__")
    )
    for mod in modules_to_include:
        if mod not in module_ignore:
            try:
                mod_name = lib_translator.get(mod, mod)
                module_versions[mod_name] = get_distribution(mod_name).version
            except:
                pass
    return module_versions


def generate_requirement_file(
    requirements: dict, file_path: str, file_name: str = "requirements.txt"
):
    """
    Generate requirements file at file_path.

    Parameters
    ----------
    requirements : dict
        Key is the library name and value is the version
    file_path : str
        Directory to save requirements.txt
    file_name : str
        Opional parameter to specify the file name
    """

    with open(os.path.join(file_path, file_name), "w") as req_file:
        for lib in requirements:
            if requirements[lib]:
                req_file.write("{}=={}\n".format(lib, requirements[lib]))
            else:
                req_file.write("{}\n".format(lib))


def _get_feature_type_and_dtype(column):
    feature_type = "unknown"
    dtype = column.dtype
    if dtype.name in ["category", "object", "boolean"]:
        feature_type = "categorical"
    elif is_numeric_dtype(dtype):
        feature_type = "continuous"
    elif is_datetime64_dtype(dtype):
        feature_type = "datetime"
    return feature_type, dtype.name


def to_dataframe(
    data: Union[
        list,
        tuple,
        pd.Series,
        np.ndarray,
        pd.DataFrame,
    ]
):
    """
    Convert to pandas DataFrame.

    Parameters
    ----------
    data: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]
        Convert data to pandas DataFrame.

    Returns
    _______
    pd.DataFrame
        pandas DataFrame.

    """
    if isinstance(data, np.ndarray) and len(data.shape) > DIMENSION:
        raise NotImplementedError(
            f"Cannot convert a numpy array with size {data.shape} to a pandas DataFrame."
        )
    if (
        isinstance(data, np.ndarray)
        or isinstance(data, list)
        or isinstance(data, tuple)
    ):
        return pd.DataFrame(data)
    elif isinstance(data, pd.Series):
        return data.to_frame()
    elif isinstance(data, dict):
        try:
            return pd.DataFrame.from_dict(data)
        except:
            raise NotImplementedError(
                "Cannot convert this dictionary to a pandas DataFrame. \
                    Check the structure to ensure it is tabular."
            )
    elif isinstance(data, pd.DataFrame):
        return data
    elif _is_dask_dataframe(data):
        return data.compute()
    else:
        raise NotImplementedError(
            f"The data type `{type(data)}` is not supported. Convert it to a pandas DataFrame."
        )


def _is_dask_dataframe(ddf):
    """
    Will determine if the given arg is a dask dataframe.
    Returns False if dask is not installed.
    """
    try:
        import dask.dataframe as dd

        return isinstance(ddf, dd.DataFrame)
    except:
        return False


def _is_dask_series(ddf):
    """
    Will determine if the given arg is a dask dataframe.
    Returns False if dask is not installed.
    """
    try:
        import dask.dataframe as dd

        return isinstance(ddf, dd.Series)
    except:
        return False


def _log_missing_module(module, package):
    """
    Log message for missing module
    """
    logger.error(f"The {module} module was not found. Install {package}.")


def _log_multivalue_feature_column_error():
    logger.error(
        "A feature column has more than one value. Only a single value is allowed."
    )


def _log_plot_high_cardinality_warning(s, length):
    logger.warning(
        f"There are too many distinct values for {s} ({length:,}) to plot. Only the top {MAX_DISPLAY_VALUES}, by cardinality, will be used."
    )


def snake_to_camel(name: str, capitalized_first_token: Optional[bool] = False) -> str:
    """Converts the snake case string to the camel representation.

    Parameters
    ----------
    name: str
        The name to convert.
    capitalized_first_token: (bool, optional). Defaults to False.
        Wether the first token needs to be capitalized or not.

    Returns
    -------
    str: The name converted to the camel representation.
    """
    tokens = name.split("_")
    return (tokens[0].capitalize() if capitalized_first_token else tokens[0]) + "".join(
        x.capitalize() if not x.isupper() else x for x in tokens[1:]
    )


def camel_to_snake(name: str) -> str:
    """Converts the camel case string to the snake representation.

    Parameters
    ----------
    name: str
        The name to convert.

    Returns
    -------
    str: The name converted to the snake representation.
    """
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def is_data_too_wide(
    data: Union[
        list,
        tuple,
        pd.Series,
        np.ndarray,
        pd.DataFrame,
    ],
    max_col_num: int,
) -> bool:
    """
    Returns true if the data has too many columns.

    Parameters
    ----------

    data: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]
        A sample of data that will be used to generate schema.
    max_col_num : int.
        The maximum column size of the data that allows to auto generate schema.
    """
    assert (
        max_col_num and isinstance(max_col_num, int) and max_col_num > 0
    ), "The parameter `max_col_num` must be a positive integer."

    data_type = type(data)
    if data_type == pd.Series or _is_dask_series(data):
        return False
    elif data_type == pd.DataFrame or _is_dask_dataframe(data):
        col_num = len(data.columns)
    elif (
        # check the column size in model_schema() after converting to pd.dataframe
        isinstance(data, np.ndarray)
        or isinstance(data, list)
        or isinstance(data, tuple)
    ):
        return False
    else:
        raise TypeError(f"The data type `{type(data)}` is not supported.")

    return col_num > max_col_num


def get_files(directory: str):
    """List out all the file names under this directory.

    Parameters
    ----------
    directory: str
        The directory to list out all the files from.

    Returns
    -------
    List
        List of the files in the directory.
    """
    directory = directory.rstrip("/")
    if os.path.exists(os.path.join(directory, ".model-ignore")):
        ignore_patterns = (
            Path(os.path.join(directory), ".model-ignore")
            .read_text()
            .strip()
            .split("\n")
        )
    else:
        ignore_patterns = []
    file_names = []
    for root, dirs, files in os.walk(directory):
        for name in files:
            file_names.append(os.path.join(root, name))
        for name in dirs:
            file_names.append(os.path.join(root, name))

    for ignore in ignore_patterns:
        if not ignore.startswith("#") and ignore.strip() != "":
            matches = []
            for file_name in file_names:
                if ignore.endswith("/"):
                    ignore = ignore[:-1] + "*"
                if not re.search(fnmatch.translate("/%s" % ignore.strip()), file_name):
                    matches.append(file_name)
            file_names = matches
    return [matched_file[len(directory) + 1 :] for matched_file in file_names]


def download_from_web(url: str, to_path: str) -> None:
    """Downloads a single file from http/https/ftp.

    Parameters
    ----------
    url : str
        The URL of the source file.
    to_path : path-like object
        Local destination path.

    Returns
    -------
    None
        Nothing
    """
    url_response = request.urlopen(url)
    with contextlib.closing(url_response) as fp:
        with open(to_path, "wb") as out_file:
            block_size = DEFAULT_BUFFER_SIZE * 8
            while True:
                block = fp.read(block_size)
                if not block:
                    break
                out_file.write(block)


def copy_from_uri(
    uri: str,
    to_path: str,
    unpack: Optional[bool] = False,
    force_overwrite: Optional[bool] = False,
    auth: Optional[Dict] = None,
) -> None:
    """Copies file(s) to local path. Can be a folder, archived folder or a separate file.
    The source files can be located in a local folder or in OCI Object Storage.

    Parameters
    ----------
    uri: str
        The URI of the source file or directory, which can be local path or
        OCI object storage URI.
    to_path: str
        The local destination path.
        If this is a directory, the source files will be placed under it.
    unpack : (bool, optional). Defaults to False.
        Indicate if zip or tar.gz file specified by the uri should be unpacked.
        This option has no effect on other files.
    force_overwrite: (bool, optional). Defaults to False.
        Whether to overwrite existing files or not.
    auth: (Dict, optional). Defaults to None.
        The default authetication is set using `ads.set_auth` API. If you need to override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
        authentication signer and kwargs required to instantiate IdentityClient object.

    Returns
    -------
    None
        Nothing

    Raises
    ------
    ValueError
        If destination path is already exist and `force_overwrite` is set to False.
    """
    if os.path.exists(to_path):
        if not force_overwrite:
            raise ValueError(
                "The destination path already exists. "
                "Set `force_overwrite` to True if you wish to overwrite."
            )
        shutil.rmtree(to_path, ignore_errors=True)

    scheme = urlparse(uri).scheme
    auth = auth or authutil.default_signer()

    with tempfile.TemporaryDirectory() as temp_dir:
        if unpack and str(uri).lower().endswith((".zip", ".tar.gz", ".gztar")):
            unpack_path = to_path
            to_path = temp_dir
        else:
            unpack_path = None
        fs = fsspec.filesystem(scheme, **auth)
        try:
            fs.get(uri, to_path, recursive=True)
        except IsADirectoryError:
            to_path = os.path.join(to_path, os.path.basename(str(uri).rstrip("/")))
            fs.get(uri, to_path, recursive=True)

        if unpack_path:
            shutil.unpack_archive(to_path, unpack_path)


def copy_file(
    uri_src: str,
    uri_dst: str,
    force_overwrite: Optional[bool] = False,
    auth: Optional[Dict] = None,
    chunk_size: Optional[int] = DEFAULT_BUFFER_SIZE,
    progressbar_description: Optional[str] = "Copying `{uri_src}` to `{uri_dst}`",
) -> str:
    """
    Copies file from `uri_src` to `uri_dst`.
    If `uri_dst` specifies a directory, the file will be copied into `uri_dst`
    using the base filename from `uri_src`.
    Returns the path to the newly created file.

    Parameters
    ----------
    uri_src: str
        The URI of the source file, which can be local path or OCI object storage URI.
    uri_dst: str
        The URI of the destination file, which can be local path or OCI object storage URI.
    force_overwrite: (bool, optional). Defaults to False.
        Whether to overwrite existing files or not.
    auth: (Dict, optional). Defaults to None.
        The default authetication is set using `ads.set_auth` API. If you need to override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
        authentication signer and kwargs required to instantiate IdentityClient object.
    chunk_size: (int, optinal). Defaults to `DEFAULT_BUFFER_SIZE`
        How much data can be copied in one iteration.

    Returns
    -------
    str
        The path to the newly created file.

    Raises
    ------
    FileExistsError
        If a destination file exists and `force_overwrite` set to `False`.
    """
    chunk_size = chunk_size or DEFAULT_BUFFER_SIZE
    auth = auth or authutil.default_signer()

    if not os.path.basename(uri_dst):
        uri_dst = os.path.join(uri_dst, os.path.basename(uri_src))
    src_path_scheme = urlparse(uri_src).scheme or "file"
    src_file_system = fsspec.filesystem(src_path_scheme, **auth)
    file_size = src_file_system.info(uri_src)["size"]

    if not force_overwrite:
        dst_path_scheme = urlparse(uri_dst).scheme or "file"
        if fsspec.filesystem(dst_path_scheme, **auth).exists(uri_dst):
            raise FileExistsError(
                f"The `{uri_dst}` exists. Please use a new file name or "
                "set force_overwrite to True if you wish to overwrite."
            )

    with fsspec.open(uri_dst, mode="wb", **auth) as fwrite:
        with fsspec.open(uri_src, mode="rb", encoding=None, **auth) as fread:
            with tqdm.wrapattr(
                fread,
                "read",
                desc=progressbar_description.format(uri_src=uri_src, uri_dst=uri_dst),
                total=file_size,
                position=0,
                leave=False,
                colour="blue",
                file=sys.stdout,
            ) as ffrom:
                while True:
                    chunk = ffrom.read(chunk_size)
                    if not chunk:
                        break
                    fwrite.write(chunk)

    return uri_dst


def remove_file(file_path: str, auth: Optional[Dict] = None) -> None:
    """
    Reoves file.

    Parameters
    ----------
    file_path: str
        The path of the source file, which can be local path or OCI object storage URI.
    auth: (Dict, optional). Defaults to None.
        The default authetication is set using `ads.set_auth` API. If you need to override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
        authentication signer and kwargs required to instantiate IdentityClient object.

    Returns
    -------
    None
        Nothing.
    """
    scheme = urlparse(file_path).scheme
    auth = auth or (scheme and authutil.default_signer()) or {}
    fs = fsspec.filesystem(scheme, **auth)
    try:
        fs.rm(file_path)
    except FileNotFoundError as e:
        raise FileNotFoundError(f"`{file_path}` not found.")
    except Exception as e:
        raise e


def folder_size(path: str) -> int:
    """Recursively calculating a size of the `path` folder.

    Parameters
    ----------
    path: str
        Path to the folder.

    Returns
    -------
    int
        The size fo the folder in bytes.
    """
    if not path:
        return 0

    if os.path.isfile(path):
        return os.path.getsize(path)

    path = os.path.join(path.rstrip("/"), "**")
    return sum(
        os.path.getsize(f) for f in glob.glob(path, recursive=True) if os.path.isfile(f)
    )


def human_size(num_bytes: int, precision: Optional[int] = 2) -> str:
    """Converts bytes size to a string representing its value in B, KB, MB and GB.

    Parameters
    ----------
    num_bytes: int
        The size in bytes.
    precision: (int, optional). Defaults to 2.
        The precision of converting the bytes value.

    Returns
    -------
    str
        A string representing the size in B, KB, MB and GB.
    """
    if not num_bytes:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    size_index = int(math.floor(math.log(num_bytes, 1024)))
    result_size = round(num_bytes / math.pow(1024, size_index), precision)
    return f"{result_size}{size_name[size_index]}"


def get_value(obj, attr, default=None):
    """Gets a copy of the value from a nested dictionary of an object with nested attributes.

    Parameters
    ----------
    obj :
        An object or a dictionary
    attr :
        Attributes as a string seprated by dot(.)
    default :
        Default value to be returned if attribute is not found.

    Returns
    -------
    Any:
        A copy of the attribute value. For dict or list, a deepcopy will be returned.

    """
    keys = attr.split(".")
    val = default
    for key in keys:
        if hasattr(obj, key):
            val = getattr(obj, key)
        elif hasattr(obj, "get"):
            val = obj.get(key, default)
        else:
            return default
        obj = val
    return copy.deepcopy(val)


def _filter_fn(adjective: str, word: str) -> bool:
    """Used to filter list of adjectives phonetically

    Parameters
    ----------
    adjective: str
        adjective word
    word: str
        word to see if should be included in list of alliterations

    Returns
    -------
    bool:
        filter or not
    """
    if adjective.startswith("f"):
        return word.startswith("f") or word.startswith("ph")
    elif adjective.startswith("q"):
        return word.startswith("q") or word.startswith("k")
    else:
        return word.startswith(adjective[0])


def get_random_name_for_resource() -> str:
    """Returns randomly generated easy to remember name. It consists from 1 adjective and 1 animal word,
    tailed by UTC timestamp (joined with '-'). This is an ADS default resource name generated for
    models, jobs, jobruns, model deployments, pipelines.

    Returns
    -------
    str
        Randomly generated easy to remember name for oci resources - models, jobs, jobruns, model deployments, pipelines.
        Example: polite-panther-2022-08-17-21:15.46; strange-spider-2022-08-17-23:55.02
    """

    adjective = random.choice(adjectives)
    animal = random.choice(
        list(filter(lambda x: _filter_fn(adjective, x), animals)) or animals
    )

    return "-".join(
        (
            adjective,
            animal,
            datetime.utcnow().strftime("%Y-%m-%d-%H:%M.%S"),
        )
    )


def batch_convert_case(spec: dict, to_fmt: str) -> Dict:
    """
    Convert the case of a dictionary of spec from camel to snake or vice versa.

    Parameters
    ----------
    spec: dict
        dictionary of spec to convert
    to_fmt: str
        format to convert to, can be "camel" or "snake"

    Returns
    -------
    dict
        dictionary of converted spec
    """
    if not spec:
        return spec

    converted = {}
    if to_fmt == "camel":
        converter = snake_to_camel
    else:
        converter = camel_to_snake
    for k, v in spec.items():
        if k == "spec":
            converted[converter(k)] = batch_convert_case(v, to_fmt)
        else:
            converted[converter(k)] = v
    return converted


def extract_region(auth: Optional[Dict] = None) -> Union[str, None]:
    """Extracts region information from the environment variables and signer.

    Parameters
    ----------
    auth: Dict
        The ADS authentication config used to initialize the client.
        Contains keys - config, signer and client_kwargs.

    Returns
    -------
    Union[str, None]
        The region identifier. For example: `us-ashburn-1`.
        Returns `None` if region cannot be extracted.
    """
    auth = auth or authutil.default_signer()

    if auth.get("config", {}).get("region"):
        return auth["config"]["region"]

    if (
        auth.get("signer")
        and hasattr(auth["signer"], "region")
        and auth["signer"].region
    ):
        return auth["signer"].region

    try:
        return json.loads(config.OCI_REGION_METADATA)["regionIdentifier"]
    except:
        pass

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents an Address feature type.

Classes:
    Address
        The Address feature type.
"""
import matplotlib.pyplot as plt
import pandas as pd
import re
from ads.feature_engineering.feature_type.string import String
from ads.feature_engineering.utils import (
    _count_unique_missing,
    _to_lat_long,
    _plot_gis_scatter,
    _zip_code,
)
from ads.feature_engineering import schema

PATTERN = re.compile(
    r"\d{1,5} [\w\s]{1,30}(?:street|st(?:\s|\.)+|avenue|ave(?:\s|\.)+|road|rd(?:\s|\.)+|highway|hwy(?:\s|\.)+|square|sq(?:\s|\.)+|trail|trl(?:\s|\.)+|drive|dr(?:\s|\.)+|court|ct(?:\s|\.)+|park|parkway|pkwy(?:\s|\.)+|circle|cir(?:\s|\.)+|boulevard|blvd(?:\s|\.)+|island|port|view|parkways)(?:suite\s?\d+|apt\.?\s?\d+|ste\.?\s?\d+)?[\w\s,]{1,30}\d{5}\W?(?=\s|$)",
    re.IGNORECASE,
)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: pd.Series
        The data to process.

    Returns
    -------
    :class:`pandas.Series`
        The logical list indicating if the data matches requirements.
    """

    def _is_address(x):
        return (
            not pd.isnull(x)
            and isinstance(x, str)
            and PATTERN.match(str(x)) is not None
        )

    return data.apply(lambda x: True if _is_address(x) else False)


class Address(String):
    """
    Type representing address.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows the location of given address on map base on zip code.

    Example
    -------
    >>> from ads.feature_engineering.feature_type.address import Address
    >>> import pandas as pd
    >>> address = pd.Series(['1 Miller Drive, New York, NY 12345',
                            '1 Berkeley Street, Boston, MA 67891',
                            '54305 Oxford Street, Seattle, WA 95132',
                            ''])
    >>> Address.validator.is_address(address)
    0     True
    1     True
    2     True
    3    False
    dtype: bool
    """

    description = "Type representing address."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count) and missing(count).

        Examples
        --------
        >>> address = pd.Series(['1 Miller Drive, New York, NY 12345',
                      '1 Berkeley Street, Boston, MA 67891',
                      '54305 Oxford Street, Seattle, WA 95132',
                      ''],
                   name='address')
        >>> address.ads.feature_type = ['address']
        >>> address.ads.feature_stat()
            Metric  Value
        0	count	4
        1	unique	3
        2	missing	1

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series provided.
        """
        return _count_unique_missing(x)

    @staticmethod
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows the location of given address on map base on zip code.

        Examples
        --------
        >>> address = pd.Series(['1 Miller Drive, New York, NY 12345',
                      '1 Berkeley Street, Boston, MA 67891',
                      '54305 Oxford Street, Seattle, WA 95132',
                      ''],
                   name='address')
        >>> address.ads.feature_type = ['address']
        >>> address.ads.feature_plot()
        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the Address feature type.
        """

        def _get_zipcode(n):
            return re.findall(r"\D(\d{5})", n)[-1]

        gis = _to_lat_long(x.loc[default_handler(x)].apply(_get_zipcode), _zip_code())
        if len(gis.index):
            return _plot_gis_scatter(gis, "longitude", "latitude")

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> address = pd.Series(['1 Miller Drive, New York, NY 12345',
                      '1 Berkeley Street, Boston, MA 67891',
                      '54305 Oxford Street, Seattle, WA 95132',
                      ''],
                   name='address')
        >>> address.ads.feature_type = ['address']
        >>> address.ads.feature_domain()
        constraints: []
        stats:
            count: 4
            missing: 1
            unique: 3
        values: Address

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the Address feature type.
        """

        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [],
        )


Address.validator.register("is_address", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from dataclasses import asdict, dataclass
from typing import Any, Tuple, Union, Dict, List

from ads.data_labeling.boundingbox import BoundingBoxItem
from ads.data_labeling.ner import NERItem


@dataclass
class Record:
    """Class representing Record.

    Attributes
    ----------
    path: str
        File path.
    content: Any
        Content of the record.
    annotation: Union[Tuple, str, List[BoundingBoxItem], List[NERItem]]
        Annotation/label of the record.
    """

    path: str = ""
    content: Any = None
    annotation: Union[Tuple, str, List[BoundingBoxItem], List[NERItem]] = None

    def to_dict(self) -> Dict:
        """Convert the Record instance to a dictionary.

        Returns
        -------
        Dict
            Dictionary representation of the Record instance.
        """
        return asdict(self)

    def to_tuple(
        self,
    ) -> Tuple[str, Any, Union[Tuple, str, List[BoundingBoxItem], List[NERItem]]]:
        """Convert the Record instance to a tuple.

        Returns
        -------
        Tuple
            Tuple representation of the Record instance.
        """
        return (self.path, self.content, self.annotation)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import inspect

import pandas as pd
from sklearn.base import TransformerMixin


def expand_lambda_function(lambda_func):
    """
    Returns a lambda function after expansion.
    """
    lambda_function = inspect.getsource(lambda_func).split(",")[1].strip()
    if lambda_function.endswith(")"):
        return lambda_function[:-1]
    return lambda_function


class DataFrameTransformer(TransformerMixin):
    """
    A DataFrameTransformer object.
    """

    def __init__(
        self, func_name, target_name, target_sample_val, args=None, kw_args=None
    ):
        self.function_name_ = func_name
        self.target_name = target_name
        self.target_sample_val = target_sample_val
        self.function_args_ = args
        self.function_kwargs_ = kw_args

    def __repr__(self):
        return "\n{}({})\n\n{}".format(
            self.function_name_,
            self.target_name,
            "\n".join(
                [
                    expand_lambda_function(f)
                    if f.__name__ == "<lambda>"
                    else "{} {}".format(f.__name__, f.__code__.co_varnames)
                    for f in self.function_args_
                    if callable(f)
                ]
            ),
        )

    def fit(self, df):
        """
        Takes in a DF and returns a fitted model
        """
        return self

    def transform(self, df):
        """
        Takes in a DF and returns a transformed DF
        """
        return self._transform(df)[0]

    def _transform(self, df):
        """
        Transform the dataframe using the function provided

        If a given function is not supported by pandas dataframe, the transform would be a no-op

        Parameters
        ----------
        df: Union[pandas.DataFrame, dask.dataframe.core.DataFrame]

        Returns
        -------
        tuple(transformed_df, is_transformed)
            transformed_df is same type as the input
            is_transformed, bool
                True if the transformation function coulfunction_args_d be applied
        """
        # add a dummy target column
        drop_target = False
        if self.target_name is not None and self.target_name not in df.columns:
            drop_target = True
            df = df.assign(**{self.target_name: self.target_sample_val})

        # check if pandas dataframe has this function
        if hasattr(df, self.function_name_):
            function = getattr(df, self.function_name_)

        # if pandas dataframe does not have this function, it is possible that the function being accessed is
        # similar to dask.dataframe.core.DataFrame.map_partitions that takes python function as the first
        # argument and applies the function on each partition. The same effect can be achieved using pipe
        elif (
            isinstance(df, pd.DataFrame)
            and len(self.function_args_) > 0
            and callable(self.function_args_[0])
        ):
            function = getattr(df, "pipe")
        else:
            # this method cannot be applied on pandas dataframe
            return df, False

        function_args = self.function_args_
        result = function(*function_args, **self.function_kwargs_)
        if drop_target:
            result = result.drop(self.target_name, axis=1)
        return result, True



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module loads and saves images from/to a local path or Oracle Object Storage.

Classes
-------
ADSImageReader
    The class reads image files from a local path
    or an Oracle Cloud Infrastructure Object Storage bucket.

Examples
--------
>>> from ads.feature_engineering import ADSImageReader
>>> image_reader = ADSImageReader.from_uri(path="oci://<bucket_name>@<namespace>/*.jpg")
>>> index=1
>>> for image in image_reader.read():
...     image.save(f"{index}.jpg")
...     index+=1
"""

import os
from io import BytesIO
from typing import Any, Dict, Generator, List, Optional, Union

import fsspec
from ads.common import auth as authutil
from PIL import Image

from .image import ADSImage
from .interface.reader import Reader


class ADSImageReader:
    """
    The class reads image files from a local path
    or an Oracle Cloud Infrastructure Object Storage bucket.

    Methods
    -------
    from_uri(cls, path: Union[str, List[str]], ...) -> ADSImageReader
        Constructs the ADSImageReader object.
    read(self) -> Generator[ADSImage, Any, Any]
        Reads images.

    Examples
    --------
    >>> from ads.feature_engineering import ADSImageReader
    >>> image_reader = ADSImageReader.from_uri(path="oci://<bucket_name>@<namespace>/*.jpg")
    >>> index=1
    >>> for image in image_reader.read():
    ...     image.save(f"{index}.jpg")
    ...     index+=1
    """

    def __init__(self, reader: Reader) -> None:
        """Initializes ADSImageReader instance.

        Parameters
        ----------
        reader: Reader
            Can be any reader implementing the ```Reader``` interface.

        Returns
        -------
        None
            Nothing.
        """
        self._reader = reader

    @classmethod
    def from_uri(
        cls,
        path: Union[str, List[str]],
        auth: Optional[Dict] = None,
    ) -> "ADSImageReader":
        """Constructs the ADSImageReader object.

        Parameters
        ----------
        path: Union[str, List[str]]
            The path where the images located.
            Can be local path, OCI object storage URI or a list of paths.
            It is also support globbing.
            Example:
                `oci://<bucket_name>@<namespace>/1.jpg`
                [`oci://<bucket_name>@<namespace>/1.jpg`, `oci://<bucket_name>@<namespace>/2.jpg`]
                `oci://<bucket_name>@<namespace>/*.jpg`

        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        ADSImageReader
            The ADSImageReader object.
        """
        auth = auth or authutil.default_signer()
        return cls(reader=ImageFileReader(path=path, auth=auth))

    def read(self) -> Generator[ADSImage, Any, Any]:
        """Read image files.

        Yields
        ------
        Generator[ADSImage, Any, Any]
        """
        return self._reader.read()


class ImageFileReader:
    """The class reads image files from a local path or an Oracle Cloud Infrastructure Object Storage bucket."""

    def __init__(
        self, path: Union[str, List[str]], auth: Optional[Dict] = None
    ) -> None:
        """Initializes ImageFileReader instance.

        Parameters
        ----------
        path: Union[str, List[str]]
            The path where the images located.
            Can be local path, OCI object storage URI or a list of paths.
            It is also support globbing.
            Example:
                `oci://<bucket_name>@<namespace>/1.jpg`
                [`oci://<bucket_name>@<namespace>/1.jpg`, `oci://<bucket_name>@<namespace>/2.jpg`]
                `oci://<bucket_name>@<namespace>/*.jpg`
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        TypeError
            If the `path` is not an instance of `str` or `List[str]`.
        ValueError
            If the `path` is not provided.
        """
        if not path:
            raise ValueError("The parameter `path` is required.")

        if not isinstance(path, (str, list)):
            raise TypeError("The `path` parameter must be a string or list of strings.")

        self.path = path
        self.auth = auth or authutil.default_signer()

    def read(self) -> Generator[ADSImage, Any, Any]:
        """Read image files.

        Yields
        ------
        Generator[ADSImage, Any, Any]
        """
        with fsspec.open_files(self.path, **self.auth) as files:
            for f in files:
                yield ADSImage(Image.open(BytesIO(f.read())), os.path.basename(f.path))



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
from abc import abstractmethod
from typing import Dict, List, Union

from ads.data_labeling.boundingbox import BoundingBoxItem
from ads.data_labeling.constants import AnnotationType, Formats
from ads.data_labeling.interface.parser import Parser
from ads.data_labeling.ner import NERItem
from ads.data_labeling.record import Record

DATASET_RECORD_INVALID_MESSAGE = "The dataset record file is invalid. "


class EntityType:
    """Entity type class for supporting multiple types of entities."""

    GENERIC = "GENERIC"
    TEXTSELECTION = "TEXTSELECTION"
    IMAGEOBJECTSELECTION = "IMAGEOBJECTSELECTION"


class RecordParser(Parser):
    """RecordParser class which parses the labels from the record.

    Examples
    --------
    >>> from ads.data_labeling.parser.export_record_parser import SingleLabelRecordParser
    >>> from ads.data_labeling.parser.export_record_parser import MultiLabelRecordParser
    >>> from ads.data_labeling.parser.export_record_parser import NERRecordParser
    >>> from ads.data_labeling.parser.export_record_parser import BoundingBoxRecordParser
    >>> import fsspec
    >>> import json
    >>> from ads.common import auth as authutil
    >>> labels = []
    >>> with fsspec.open("/path/to/records_file.jsonl", **authutil.api_keys()) as f:
    >>>     for line in f:
    >>>         bounding_box_labels = BoundingBoxRecordParser("source_data_path").parse(json.loads(line))
    >>>         labels.append(bounding_box_labels)
    """

    def __init__(
        self, dataset_source_path: str, format: str = None, categories: List[str] = None
    ) -> "RecordParser":
        """Initiates a RecordParser instance.

        Parameters
        ----------
        dataset_source_path: str
            Dataset source path.
        format: (str, optional). Defaults to None.
            Output format of annotations.
        categories: (List[str], optional). Defaults to None.
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']

        Returns
        -------
        RecordParser
            RecordParser instance.
        """
        self.dataset_source_path = dataset_source_path
        self.format = format
        self.categories = categories

    def parse(self, record: Dict) -> "Record":
        """Extracts the annotations from the record content.
        Constructs and returns a Record instance containing the file path and the labels.

        Parameters
        ----------
        record: Dict
            Content of the record from the record file.

        Returns
        -------
        Record
            Record instance which contains the file path as well as the annotations.
        """
        return Record(
            path=self.dataset_source_path + record["sourceDetails"]["path"],
            annotation=self._extract_annotations(
                record=record, format=self.format, categories=self.categories
            ),
        )

    @abstractmethod
    def _extract_annotations(
        self,
        record: Dict,
        **kwargs,
    ) -> Union[str, List[str], List[BoundingBoxItem], List[NERItem]]:
        """Extracts annotations from the record content. Each Parser class
        needs to implement this function.

        Parameters
        ----------
        record: Dict
            Content of the record from the record file.
        kwargs: Dict
            format: str
                Output format of annotations. Can be "spacy" or "yolo".

        Returns
        -------
        Union[str, List[str], List[BoundingBoxItem], List[NERItem]]
            Label(s).
        """
        pass

    def _validate(self, record: Dict) -> None:
        """Validates the record to ensure it contains certain fields.

        Raises
        ------
        ValueError
            If record format is incorrect.
        """
        if (
            "annotations" not in record
            or not isinstance(record["annotations"], list)
            or "entities" not in record["annotations"][0]
            or not isinstance(record["annotations"][0]["entities"], list)
            or "entityType" not in record["annotations"][0]["entities"][0]
            or "labels" not in record["annotations"][0]["entities"][0]
        ):
            raise ValueError(
                f"{DATASET_RECORD_INVALID_MESSAGE}"
                "At least one record is in the wrong format. "
                "Use the `DataLabeling.export()` method to create a new dataset record file."
            )


class SingleLabelRecordParser(RecordParser):
    """SingleLabelRecordParser class which parses the label of Single label data."""

    def _extract_annotations(self, record: Dict, **kwargs) -> Union[str, None]:
        """Extract the labels of the single label annotation class.

        Parameters
        ----------
        record: Dict
            Content of the record from the record file.

        Returns
        -------
        Union[str, None]
            A label or None for the unlabeled record.
        """
        if "annotations" in record:
            self._validate(record)
            return record["annotations"][0]["entities"][0]["labels"][0]["label_name"]
        else:
            return None

    def _validate(self, record: Dict) -> None:
        """Validates the format of the single label record.

        Raises
        ------
        ValueError
            If record format is incorrect.
        """
        super()._validate(record)
        if record["annotations"][0]["entities"][0]["entityType"] != EntityType.GENERIC:
            raise ValueError(
                f"{DATASET_RECORD_INVALID_MESSAGE}"
                "At least one record contains the invalid entity type:  "
                f"`{record['annotations'][0]['entities'][0]['entityType']}`. The entity "
                f"type of the Single Label annotation must be `{EntityType.GENERIC}`. "
                "Use the `DataLabeling.export()` method to create a new dataset record file."
            )
        if len(record["annotations"][0]["entities"][0]["labels"]) != 1:
            raise ValueError(
                f"{DATASET_RECORD_INVALID_MESSAGE}"
                "At least one record contains an invalid number of records: "
                f"{len(record['annotations'][0]['entities'][0]['labels'])}. "
                "The Single Label annotation expects only one label for each record. "
                "Use the `DataLabeling.export()` method to create a new dataset record file."
            )


class MultiLabelRecordParser(RecordParser):
    """MultiLabelRecordParser class which parses the label of Multiple label data."""

    def _extract_annotations(self, record: Dict, **kwargs) -> Union[List[str], None]:
        """Extract labels of the Multi label annotation class.

        Parameters
        ----------
        record: Dict
            Content of the record from the record file.

        Returns
        -------
        Union[List[str], None]
            List of labels or None for the unlabeled record.
        """
        if "annotations" in record:
            self._validate(record)
            return [
                label["label_name"]
                for label in record["annotations"][0]["entities"][0]["labels"]
            ]
        else:
            return None

    def _validate(self, record: Dict) -> None:
        """Validates the format of the multi label record.

        Raises
        ------
        ValueError
            If record format is incorrect.
        """
        super()._validate(record)
        if record["annotations"][0]["entities"][0]["entityType"] != EntityType.GENERIC:
            raise ValueError(
                f"At least one of the dataset records contains the invalid entity type: "
                f"`{record['annotations'][0]['entities'][0]['entityType']}`. "
                f"The entity type of the Multi Label annotation must be `{EntityType.GENERIC}`."
            )
        if len(record["annotations"][0]["entities"][0]["labels"]) < 1:
            raise ValueError(
                f"At least one of the dataset records contains invalid number of labels: "
                f"`{len(record['annotations'][0]['entities'][0]['labels'])}`. "
                f"The Multi Label annotation expects at least one label for each record."
            )


class NERRecordParser(RecordParser):
    """NERRecordParser class which parses the label of NER label data."""

    def _extract_annotations(
        self, record: Dict, **kwargs
    ) -> Union[List[NERItem], None]:
        """Extracts the labels of the NER annotation class.

        Parameters
        ----------
        record: Dict
            Content of the record from the record file.
        kwargs: Dict
            format: str
                Output format of annotations. Can be "spacy" or None.
                When None, it outputs List[NERItem]. When "spacy", it
                outputs List[Tuple].

        Returns
        -------
        Union[List[NERItem], List[Tuple], None]
            The list of NERItem objects or list of tuples in spacy format.
        """
        if "annotations" in record:
            self._validate(record)
            format = kwargs.get("format", None)
            items = []
            for entity in record["annotations"][0]["entities"]:
                label = entity["labels"][0]["label_name"]
                offset = entity["textSpan"]["offset"]
                length = entity["textSpan"]["length"]
                item = NERItem(label=label, offset=offset, length=length)
                if (
                    format
                    and isinstance(format, str)
                    and format.lower() == Formats.SPACY
                ):
                    item = item.to_spacy()
                items.append(item)
            return items
        else:
            return None

    def _validate(self, record: Dict) -> None:
        """Validates the format of the NER label record.

        Raises
        ------
        ValueError
            If record format is incorrect.
        """
        super()._validate(record)
        if (
            record["annotations"][0]["entities"][0]["entityType"]
            != EntityType.TEXTSELECTION
        ):
            raise ValueError(
                f"{DATASET_RECORD_INVALID_MESSAGE}"
                "At least one record contains the invalid entity type:  "
                f"`{record['annotations'][0]['entities'][0]['entityType']}`. The entity type "
                f"of the Single Label annotation must be `{EntityType.TEXTSELECTION}`. "
                "Use the `DataLabeling.export()` method to create a new dataset record file."
            )
        if os.path.splitext(record["sourceDetails"]["path"])[1].lower() != ".txt":
            raise ValueError(
                f"The file ({record['sourceDetails']['path']}) must be a text file and have a '.txt' file extension."
            )


class BoundingBoxRecordParser(RecordParser):
    """BoundingBoxRecordParser class which parses the label of BoundingBox label data."""

    def _extract_annotations(
        self, record: Dict, **kwargs: Dict
    ) -> Union[List[BoundingBoxItem], None]:
        """Extracts the labels of the Object Detection annotation class.

        Parameters
        ----------
        record: Dict
            Content of the record from the record file.
        kwargs: Dict
            format: str
                Output format of annotations. Can be None or "yolo".
                When None, it outputs List[BoundingBoxItem]. When "yolo", it
                outputs List[List[Tuple]].
            categories: Optional List[str]
                 The list of object categories in proper order for model training.
                 Only used when bounding box annotations are in YOLO format.

        Returns
        -------
        Union[List[BoundingBoxItem], List[List[Tuple]], None]
            The list of BoundingBoxItem objects or list of tuples in YOLO format.
        """
        if not "annotations" in record:
            return None

        self._validate(record)
        format = kwargs.get("format", None)
        categories = kwargs.get("categories", None)
        items = []
        for entity in record["annotations"][0]["entities"]:
            labels = [label["label_name"] for label in entity["labels"]]
            coords = entity["boundingPolygon"]["normalizedVertices"]
            top_left = (float(coords[0]["x"]), float(coords[0]["y"]))
            bottom_left = (float(coords[1]["x"]), float(coords[1]["y"]))
            bottom_right = (float(coords[2]["x"]), float(coords[2]["y"]))
            top_right = (float(coords[3]["x"]), float(coords[3]["y"]))
            item = BoundingBoxItem(
                labels=labels,
                bottom_left=bottom_left,
                top_left=top_left,
                top_right=top_right,
                bottom_right=bottom_right,
            )
            if format and isinstance(format, str) and format.lower() == Formats.YOLO:
                item = item.to_yolo(categories=categories)
            items.append(item)

        return items

    def _validate(self, record: Dict) -> None:
        """Validates the format of the image label record.

        Raises
        ------
        ValueError
            If record format is incorrect.
        """
        super()._validate(record)
        if (
            record["annotations"][0]["entities"][0]["entityType"]
            != EntityType.IMAGEOBJECTSELECTION
        ):
            raise ValueError(
                f"{DATASET_RECORD_INVALID_MESSAGE}"
                "At least one record contains the invalid entity type:  "
                f"`{record['annotations'][0]['entities'][0]['entityType']}`. The entity type "
                f"of the Single Label annotation must be `{EntityType.IMAGEOBJECTSELECTION}`. "
                "Use the `DataLabeling.export()` method to create a new dataset record file."
            )
        if os.path.splitext(record["sourceDetails"]["path"])[1].lower() not in [
            ".jpg",
            ".png",
            ".jpeg",
        ]:
            raise ValueError(
                f"The file ({record['sourceDetails']['path']}) must be a jpg, jpeg or png file."
            )


class RecordParserFactory:
    """RecordParserFactory class which contains a list of registered parsers
    and allows to register new RecordParsers.

    Current parsers include:
        * SingleLabelRecordParser
        * MultiLabelRecordParser
        * NERRecordParser
        * BoundingBoxRecordParser
    """

    _parsers = {
        AnnotationType.SINGLE_LABEL: SingleLabelRecordParser,
        AnnotationType.MULTI_LABEL: MultiLabelRecordParser,
        AnnotationType.ENTITY_EXTRACTION: NERRecordParser,
        AnnotationType.BOUNDING_BOX: BoundingBoxRecordParser,
    }

    @staticmethod
    def parser(
        annotation_type: str,
        dataset_source_path: str,
        format: str = None,
        categories: List[str] = None,
    ) -> "RecordParser":
        """Gets the parser based on the annotation_type.

        Parameters
        ----------
        annotation_type: str
            Annotation type which can be SINGLE_LABEL, MULTI_LABEL, ENTITY_EXTRACTION
            and BOUNDING_BOX.
        dataset_source_path: str
            Dataset source path.
        format: (str, optional). Defaults to None.
            Output format of annotations. Can be None, "spacy" for dataset
            Entity Extraction type or "yolo" for Object Detection type.
            When None, it outputs List[NERItem] or List[BoundingBoxItem].
            When "spacy", it outputs List[Tuple].
            When "yolo", it outputs List[List[Tuple]].
        categories: (List[str], optional). Defaults to None.
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']

        Returns
        -------
        RecordParser
            RecordParser corresponding to the annotation type.

        Raises
        ------
        ValueError
            If annotation_type is not supported.
        """

        if not annotation_type in RecordParserFactory._parsers:
            raise ValueError(
                f"The {annotation_type} is not supported. Choose from "
                f"`{AnnotationType.SINGLE_LABEL}`, `{AnnotationType.MULTI_LABEL}`, "
                f"`{AnnotationType.ENTITY_EXTRACTION}` and `{AnnotationType.BOUNDING_BOX}`."
            )

        return RecordParserFactory._parsers[annotation_type](
            dataset_source_path=dataset_source_path,
            format=format,
            categories=categories,
        )

    @classmethod
    def register(cls, annotation_type: str, parser) -> None:
        """Registers a new parser.

        Parameters
        ----------
        annotation_type: str
            Annotation type which can be SINGLE_LABEL, MULTI_LABEL, ENTITY_EXTRACTION
            and BOUNDING_BOX.
        parser: RecordParser
            A new Parser class to be registered.

        Returns
        -------
        None
            Nothing.
        """
        cls._parsers[annotation_type] = parser
